"""Test SWAN boundary components."""

import pytest

# Import test utilities
from test_utils.logging import get_test_logger

# Initialize logger
logger = get_test_logger(__name__)

from rompy_swan.components.boundary import (
    BOUNDNEST1,
    BOUNDNEST2,
    BOUNDNEST3,
    BOUNDSPEC,
    INITIAL,
)
from rompy_swan.subcomponents.boundary import (
    CONSTANTPAR,
    HOTMULTIPLE,
    HOTSINGLE,
    PAR,
    SIDE,
    ZERO,
)
from rompy_swan.subcomponents.spectrum import JONSWAP, SHAPESPEC


def test_initial_default():
    par = INITIAL()
    assert par.render() == "INITIAL DEFAULT"


def test_initial_zero():
    par = INITIAL(kind=ZERO())
    assert par.render() == "INITIAL ZERO"


def test_initial_par():
    par = INITIAL(
        kind=PAR(
            hs=1.0,
            per=10.0,
            dir=0.0,
            dd=10.0,
        ),
    )
    assert par.render() == "INITIAL PAR hs=1.0 per=10.0 dir=0.0 dd=10.0"


def test_initial_hotsingle():
    par = INITIAL(
        kind=HOTSINGLE(
            fname="hotfile.txt",
            format="unformatted",
        ),
    )
    assert par.render() == "INITIAL HOTSTART SINGLE fname='hotfile.txt' UNFORMATTED"


def test_initial_hotmultiple():
    par = INITIAL(
        kind=HOTMULTIPLE(
            fname="hotfile.txt",
            format="unformatted",
        ),
    )
    assert par.render() == "INITIAL HOTSTART MULTIPLE fname='hotfile.txt' UNFORMATTED"


def test_boundspec():
    bnd = BOUNDSPEC(
        shapespec=SHAPESPEC(
            shape=JONSWAP(gamma=3.3),
            per_type="peak",
            dspr_type="power",
        ),
        location=SIDE(
            side="west",
        ),
        data=CONSTANTPAR(
            hs=1.0,
            per=10.0,
            dir=0.0,
            dd=10.0,
        ),
    )
    assert "BOUND SHAPESPEC JONSWAP gamma=3.3 PEAK DSPR POWER" in bnd.render()
    assert "BOUNDSPEC SIDE WEST CCW CONSTANT PAR hs=1.0" in bnd.render()


def test_boundnest1():
    bnd = BOUNDNEST1(fname="boundary_swan.txt")
    assert bnd.render() == "BOUNDNEST1 NEST fname='boundary_swan.txt' CLOSED"
    bnd = BOUNDNEST1(fname="boundary_swan.txt", rectangle="open")
    assert bnd.render() == "BOUNDNEST1 NEST fname='boundary_swan.txt' OPEN"


def test_boundnest2_format():
    bnd = BOUNDNEST2(fname="boundary_wam.txt", format="free")
    assert bnd.render() == "BOUNDNEST2 WAMNEST fname='boundary_wam.txt' FREE lwdate=12"
    for fmt in ["cray", "wkstat"]:
        bnd = BOUNDNEST2(fname="boundary_wam.txt", format=fmt)
        assert bnd.render().endswith(f"UNFORMATTED {fmt.upper()} lwdate=12")
    with pytest.raises(ValueError):
        BOUNDNEST2(fname="boundary_wam.txt", format="unknown")


def test_boundnest2_xygc():
    bnd = BOUNDNEST2(fname="boundary_wam.txt", format="free", xgc=0.0, ygc=0.0)
    assert bnd.render().endswith("FREE xgc=0.0 ygc=0.0 lwdate=12")


def test_boundnest2_lwdate():
    for lwdate in [10, 12, 14]:
        bnd = BOUNDNEST2(fname="boundary_wam.txt", format="free", lwdate=lwdate)
        assert (
            bnd.render()
            == f"BOUNDNEST2 WAMNEST fname='boundary_wam.txt' FREE lwdate={lwdate}"
        )
    with pytest.raises(ValueError):
        BOUNDNEST2(fname="boundary_wam.txt", format="wkstat", lwdate=11)


def test_boundnest3_format():
    for fmt in ["free", "unformatted"]:
        bnd = BOUNDNEST3(fname="bnd_ww3.txt", format=fmt)
    assert bnd.render() == f"BOUNDNEST3 WW3 fname='bnd_ww3.txt' {fmt.upper()} CLOSED"
    with pytest.raises(ValueError):
        BOUNDNEST3(fname="bnd_ww3.txt", format="unknown")


def test_boundnest3_xygc():
    bnd = BOUNDNEST3(fname="bnd_ww3.txt", format="free", xgc=0.0, ygc=0.0)
    assert bnd.render().endswith("FREE CLOSED xgc=0.0 ygc=0.0")



"""Test subcomponents."""

import pytest
from pydantic import ValidationError

# Import test utilities
from test_utils.logging import get_test_logger

# Initialize logger
logger = get_test_logger(__name__)

from rompy_swan.subcomponents.boundary import (
    CONSTANTFILE,
    CONSTANTPAR,
    SEGMENT,
    SIDE,
    VARIABLEFILE,
    VARIABLEPAR,
)


def test_side():
    side = SIDE(side="north")
    assert side.render().strip() == "SIDE NORTH CCW"
    with pytest.raises(ValidationError):
        SIDE(side="wrong_dir")


def test_segment_xy():
    seg = SEGMENT(points=dict(model_type="xy", x=[0, 1, 2], y=[0, 1, 2]))
    assert seg.render().startswith("SEGMENT XY")


def test_segment_ij():
    seg = SEGMENT(points=dict(model_type="ij", i=[0, 1, 2], j=[0, 1, 2]))
    assert seg.render().startswith("SEGMENT IJ")


def test_par_constant():
    par = CONSTANTPAR(hs=1.0, per=10.0, dir=0.0, dd=0.0)
    assert par.render() == "CONSTANT PAR hs=1.0 per=10.0 dir=0.0 dd=0.0"


def test_par_variable():
    par = VARIABLEPAR(
        hs=[1.0, 1.0], per=[10.0, 10.0], dir=[0.0, 0.0], dd=[0.0, 0.0], len=[0, 1]
    )
    assert par.render().startswith("VARIABLE PAR &")


def test_varpar_all_same_size():
    with pytest.raises(ValidationError):
        VARIABLEPAR(hs=[1.0], per=[10.0], dir=[0.0], dd=[0.0, 0.0], len=[0.0])


def test_file_constant():
    file = CONSTANTFILE(fname="tpar.txt")
    assert file.render() == "CONSTANT FILE fname='tpar.txt'"


def test_file_variable():
    file = VARIABLEFILE(fname=["tpar0.txt", "tpar1.txt"], len=[0, 1])
    assert file.render().startswith("VARIABLE FILE &")


def test_varfile_all_same_size():
    with pytest.raises(ValidationError):
        VARIABLEFILE(fname=["tpar0.txt", "tpar1.txt"], len=[0, 1, 2])



"""
SWAN Startup Components

This module contains components for initializing and configuring SWAN model runs,
including project settings, coordinate systems, and run modes.
"""

from typing import Literal, Optional

from pydantic import Field, field_validator

from rompy.logging import get_logger
from rompy_swan.components.base import BaseComponent
from rompy_swan.subcomponents.startup import CARTESIAN, SPHERICAL

logger = get_logger(__name__)


class PROJECT(BaseComponent):
    """SWAN Project.

    .. code-block:: text

        PROJECT 'name' 'nr' 'title' 'title2 'title3'

    With this required command the user defines a number of strings to identify the
    SWAN run (project name e.g., an engineering project) in the print and plot file.

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.components.startup import PROJECT
        proj = PROJECT(nr="01")
        print(proj.render())
        proj = PROJECT(
            name="waus",
            nr="001",
            title1="Western Australia",
            title2="Perth Nest"
        )
        print(proj.render())

    """

    model_type: Literal["project", "PROJECT"] = Field(
        default="project",
        description="Model type discriminator",
    )
    name: Optional[str] = Field(
        default=None,
        description="Is the name of the project, at most 16 characters long",
        max_length=16,
    )
    nr: str = Field(
        description=(
            "Is the run identification (to be provided as a character string; e.g. "
            "the run number) to distinguish this run among other runs for the same "
            "project; it is at most 4 characters long. It is the only required "
            "information in this command."
        ),
        max_length=4,
    )
    title1: Optional[str] = Field(
        default=None,
        description=(
            "A string of at most 72 characters provided by the user to appear in the "
            "output of the program for the user's convenience (SWAN default: blanks)"
        ),
        max_length=72,
    )
    title2: Optional[str] = Field(
        default=None, description="Same as 'title1'", max_length=72
    )
    title3: Optional[str] = Field(
        default=None, description="Same as 'title1'", max_length=72
    )

    def cmd(self) -> str:
        repr = "PROJECT"
        if self.name is not None:
            repr += f" name='{self.name}'"
        repr += f" nr='{self.nr}'"
        if self.title1 is not None:
            repr += f" title1='{self.title1}'"
        if self.title2 is not None:
            repr += f" title2='{self.title2}'"
        if self.title3 is not None:
            repr += f" title3='{self.title3}'"
        return repr


class SET(BaseComponent):
    """SWAN setting commands.

    .. code-block:: text

        SET [level] [nor] [depmin] [maxmes] [maxerr] [grav] [rho] [cdcap] &
            [inrhog] [hsrerr] NAUTICAL|->CARTESIAN [pwtail] [froudmax] [icewind]

    With this optional command the user assigns values to various general parameters.

    Notes
    -----
    The error level `maxerr` is coded as follows:

    * 1: warnings
    * 2: errors (possibly automatically repaired or repairable by SWAN)
    * 3: severe errors

    Default values for `pwtail` depend on formulations of physics:

    * command GEN1: `pwtail = 5`
    * command GEN2: `pwtail = 5`
    * command GEN3 KOMEN: `pwtail = 4`
    * command GEN3 WESTH: `pwtail = 4`
    * command GEN3 JANSSEN: `pwtail = 5`

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.components.startup import SET
        set = SET(level=0.5, direction_convention="nautical")
        print(set.render())
        set = SET(
            level=-1.0,
            nor=90,
            depmin=0.01,
            maxerr=3,
            grav=9.81,
            rho=1025,
            cdcap=2.5e-3,
            inrhog=0,
            hsrerr=0.1,
            direction_convention="nautical",
        )
        print(set.render())

    """

    model_type: Literal["set", "SET"] = Field(
        default="set", description="Model type discriminator"
    )
    level: Optional[float] = Field(
        default=None,
        description=(
            "Increase in water level that is constant in space and time can be given "
            "with this option, `level` is the value of this increase (in m). For a "
            "variable water level reference is made to the commands "
            "INPGRID and READINP (SWAN default: 0)"
        ),
        examples=[0],
    )
    nor: Optional[float] = Field(
        default=None,
        description=(
            "Direction of North with respect to the x-axis (measured "
            "counterclockwise); default `nor = 90`, i.e. x-axis of the problem "
            "coordinate system points East. When spherical coordinates are used "
            "(see command COORD) the value of `nor` may not be modified"
        ),
        ge=-360.0,
        le=360.0,
    )
    depmin: Optional[float] = Field(
        default=None,
        description=(
            "Threshold depth (in m). In the computation any positive depth smaller "
            "than `depmin` is made equal to `depmin` (SWAN default: 0.05)"
        ),
        ge=0.0,
    )
    maxmes: Optional[int] = Field(
        default=None,
        description=(
            "Maximum number of error messages during the computation at which the "
            "computation is terminated. During the computational process messages are "
            "written to the print file (SWAN default: 200)"
        ),
        ge=0,
    )
    maxerr: Optional[Literal[1, 2, 3]] = Field(
        default=None,
        description=(
            "During pre-processing SWAN checks input data. Depending on the severity "
            "of the errors encountered during this pre-processing, SWAN does not "
            "start computations. The user can influence the error level above which "
            "SWAN will  not start computations (at the level indicated the "
            "computations will continue) (SWAN default: 1)"
        ),
    )
    grav: Optional[float] = Field(
        default=None,
        description="The gravitational acceleration (in m/s2) (SWAN default: 9.81)",
        ge=0.0,
    )
    rho: Optional[float] = Field(
        default=None,
        description="The water density (in kg/m3) (SWAN default: 1025)",
        ge=0.0,
    )
    cdcap: Optional[float] = Field(
        default=None,
        description=(
            "The maximum value for the wind drag coefficient. A value of 99999 means"
            "no cutting off the drag coefficient. A suggestion for this parameter is "
            "`cdcap = 2.5x 10-3` (SWAN default: 99999) "
        ),
        ge=0.0,
    )
    inrhog: Optional[Literal[0, 1]] = Field(
        default=None,
        description=(
            "To indicate whether the user requires output based on variance or based "
            "on true energy (see Section 2.5). `inrhog` = 0: output based on variance, "
            "`inrhog` = 1: output based on true energy (SWAN default: 0)"
        ),
    )
    hsrerr: Optional[float] = Field(
        default=None,
        description=(
            "The relative difference between the user imposed significant wave height "
            "and the significant wave height computed by SWAN (anywhere along the "
            "computational grid boundary) above which a warning will be given. This "
            "relative difference is the difference normalized with the user provided "
            "significant wave height. This warning will be given for each boundary "
            "grid point where the problem occurs (with its x- and y-index number of "
            "the computational grid). The cause of the difference is explained in "
            "Section 2.6.3. To suppress these warnings (in particular for "
            "nonstationary computations), set `hsrerr` at a very high value or use "
            "command OFF BNDCHK (SWAN default: 0.10) (ONLY MEANT FOR STRUCTURED GRIDS)"
        ),
        ge=0.0,
    )
    direction_convention: Literal["nautical", "cartesian"] = Field(
        description=(
            "Direction convention: `nautical` indicates that the Nautical convention "
            "for wind and wave direction (SWAN input and output) will be used, "
            "`cartesian` indicates that the Cartesian convention for wind and wave "
            "direction will be used. For definition, see Section 2.5 or Appendix A "
            "(SWAN default: `cartesian`)"
        ),
    )
    pwtail: Optional[int] = Field(
        default=None,
        description=(
            "Power of high frequency tail; defines the shape of the spectral tail "
            "above the highest prognostic frequency `fhigh` (see command CGRID). "
            "The energy density is assumed to be proportional to frequency to the "
            "power `pwtail`. If the user wishes to use another value, then this SET "
            "command should be located in the command file after GEN1, GEN2 or GEN3 "
            "command (these will override the SET command with respect to `pwtail`)"
        ),
        ge=0,
    )
    froudmax: Optional[float] = Field(
        default=None,
        description=(
            "Is the maximum Froude number (`U/√gd` with `U` the current and `d` the "
            "water depth). The currents taken from a circulation model may mismatch "
            "with given water depth `d` in the sense that the Froude number becomes "
            "larger than 1. For this, the current velocities will be maximized by "
            "Froude number times `sqrt(gh)` (SWAN default: 0.8)"
        ),
        ge=0.0,
    )
    icewind: Optional[Literal[0, 1]] = Field(
        default=None,
        description=(
            "Controls the scaling of wind input by open water fraction. Default value "
            "of zero corresponds to the case where wind input is scaled by the open "
            "water fraction. If `icewind = 1` then sea ice does not affect wind input "
            "directly. (Though there is still indirect effect via the sea ice sink "
            "term; see command SICE) (SWAN default: 0)"
        ),
    )

    @field_validator("pwtail")
    @classmethod
    def pwtail_after_gen(cls, v):
        if v is not None:
            logger.warning("pwtail only has effect if set after GEN command")
        return v

    def cmd(self) -> str:
        repr = "SET"
        if self.level is not None:
            repr += f" level={self.level}"
        if self.nor is not None:
            repr += f" nor={self.nor}"
        if self.depmin is not None:
            repr += f" depmin={self.depmin}"
        if self.maxmes is not None:
            repr += f" maxmes={self.maxmes}"
        if self.maxerr is not None:
            repr += f" maxerr={self.maxerr}"
        if self.grav is not None:
            repr += f" grav={self.grav}"
        if self.rho is not None:
            repr += f" rho={self.rho}"
        if self.cdcap is not None:
            repr += f" cdcap={self.cdcap}"
        if self.inrhog is not None:
            repr += f" inrhog={self.inrhog}"
        if self.hsrerr is not None:
            repr += f" hsrerr={self.hsrerr}"
        if self.direction_convention is not None:
            repr += f" {self.direction_convention.upper()}"
        if self.pwtail is not None:
            repr += f" pwtail={self.pwtail}"
        if self.froudmax is not None:
            repr += f" froudmax={self.froudmax}"
        if self.icewind is not None:
            repr += f" icewind={self.icewind}"
        return repr


class MODE(BaseComponent):
    """SWAN Mode.

    .. code-block:: text

        MODE ->STATIONARY|NONSTATIONARY ->TWODIMENSIONAL|ONEDIMENSIONAL

    With this optional command the user indicates that the run will be either
    stationary or nonstationary and one-dimensional (1D-mode) or two-dimensional
    (2D-mode). Nonstationary means either (see command COMPUTE):

    * (a) one nonstationary computations or
    * (b) a sequence of stationary computations or
    * (c) a mix of (a) and (b)

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.components.startup import MODE
        mode = MODE()
        print(mode.render())
        mode = MODE(kind="nonstationary", dim="twodimensional")
        print(mode.render())

    """

    model_type: Literal["mode", "MODE"] = Field(
        default="mode", description="Model type discriminator."
    )
    kind: Literal["stationary", "nonstationary"] = Field(
        default="stationary",
        description="Indicates if run will be stationary or nonstationary",
    )
    dim: Literal["onedimensional", "twodimensional"] = Field(
        default="twodimensional",
        description=(
            "Indicates that the run will be either one-dimensional (1D-mode) or "
            "two-dimensional (2D-mode)"
        ),
    )

    def cmd(self) -> str:
        return f"MODE {self.kind.upper()} {self.dim.upper()}"


class COORDINATES(BaseComponent):
    """SWAN Coordinates.

    .. code-block:: text

        COORDINATES ->CARTESIAN|SPHERICAL REPEATING

    Command to choose between Cartesian and spherical coordinates (see Section 2.5).
    A nested SWAN run must use the same coordinate system as the coarse grid SWAN run.

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.components.startup import COORDINATES
        coords = COORDINATES()
        print(coords.render())
        coords = COORDINATES(
            kind=dict(model_type="spherical", projection="ccm"),
            reapeating=True,
        )
        print(coords.render())

    """

    model_type: Literal["coordinates", "COORDINATES"] = Field(
        default="coordinates",
        description="Model type discriminator",
    )
    kind: CARTESIAN | SPHERICAL = Field(
        default_factory=CARTESIAN,
        description="Coordinates kind",
    )
    reapeating: bool = Field(
        default=False,
        description=(
            "This option is only for academic cases. It means that wave energy "
            "leaving at one end of the domain (in computational x-direction) enter at "
            "the other side; it is as if the wave field repeats itself in x-direction "
            "with the length of the domain in x-direction. This option cannot be used "
            "in combination with computation of set-up (see command SETUP). This "
            "option is available only with regular grids"
        ),
    )

    def cmd(self) -> str:
        repr = f"COORDINATES {self.kind.render()}"
        if self.reapeating:
            repr += " REPEATING"
        return repr



"""
SWAN Numerics Components

This module contains components for configuring the numerical schemes and parameters
used in SWAN for wave propagation, frequency shifting, and other numerical aspects.
"""

from typing import Annotated, Literal, Optional, Union

from pydantic import Field

from rompy.logging import get_logger
from rompy_swan.components.base import BaseComponent
from rompy_swan.subcomponents.numerics import (
    ACCUR,
    BSBT,
    CSIGMA,
    CTHETA,
    DIRIMPL,
    GSE,
    SETUP,
    SIGIMPL,
    STOPC,
)

logger = get_logger(__name__)


PROP_TYPE = Annotated[
    Union[BSBT, GSE],
    Field(description="Propagation scheme", discriminator="model_type"),
]


class PROP(BaseComponent):
    """Propagation scheme.

    .. code-block:: text

        PROP BSTB|GSE

    Notes
    -----
    * The scheme defaults to `S&L` and `SORDUP` for nonstationary and stationary
      simulations if not specified.
    * All schemes (BSBT, SORDUP and S&L) can be used in combination with curvilinear
      grids. With the higher order schemes (S&L and SORDUP) it is important to use a
      gradually varying grid otherwise there may be a severe loss of accuracy. If sharp
      transitions in the grid cannot be avoided it is safer to use the BSBT scheme.
    * In the computation with unstructured meshes, a lowest order upwind scheme will be
      employed. This scheme is very robust but rather diffusive. This may only be
      significant for the case when swell waves propagate over relative large distances
      (in the order of thousands of kilometers) within the model domain. However and
      most fortunately, in such a case this will alleviate the garden-sprinkler effect.
    * Alleviating the garden-sprinkler effect by adding some diffusion makes the SWAN
      computation conditionally stable. You can either use (i) a smaller time step,
      (ii) a lower value of `waveage`, (iii) better resolution in the directional
      space, or (iv) worse resolution in the geographic space, in order of preference,
      to make the model stable when necessary.

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.components.numerics import PROP
        prop = PROP()
        print(prop.render())
        prop = PROP(scheme=dict(model_type="bsbt"))
        print(prop.render())
        prop = PROP(
            scheme=dict(
                model_type="gse",
                waveage=dict(delt="PT5H", dfmt="hr"),
            ),
        )
        print(prop.render())

    """

    model_type: Literal["prop", "PROP"] = Field(
        default="prop", description="Model type discriminator"
    )
    scheme: Optional[PROP_TYPE] = Field(
        default=None,
        description=(
            "Propagation scheme, by default S&L for nonstationary and SORDUP for "
            "stationary computation."
        ),
    )

    def cmd(self) -> str:
        """Command file string for this component."""
        repr = "PROP"
        if self.scheme is not None:
            repr += f" {self.scheme.render()}"
        return repr


class NUMERIC(BaseComponent):
    """Numerical properties.

    .. code-block:: text

        NUMeric ( STOPC [dabs] [drel] [curvat] [npnts] ->STAT|NONSTAT [limiter] ) &
            ( DIRimpl [cdd] ) ( SIGIMpl [css] [eps2] [outp] [niter] ) &
            ( CTheta [cfl] ) ( CSigma [cfl] ) ( SETUP [eps2] [outp] [niter] )

    Examples
    --------
    .. ipython:: python
        :okwarning:

        from rompy_swan.components.numerics import NUMERIC
        numeric = NUMERIC()
        print(numeric.render())
        numeric = NUMERIC(
            stop=dict(
                model_type="stopc",
                dabs=0.05,
                drel=0.01,
                curvat=0.05,
                npnts=99.5,
            ),
            dirimpl=dict(cdd=0.5),
            sigimpl=dict(css=0.5, eps2=1e-4, outp=0, niter=20),
            ctheta=dict(cfl=0.9),
            csigma=dict(cfl=0.9),
            setup=dict(eps2=1e-4, outp=0, niter=20),
        )
        print(numeric.render())

    """

    model_type: Literal["numeric", "NUMERIC"] = Field(
        default="numeric", description="Model type discriminator"
    )
    stop: Optional[Union[STOPC, ACCUR]] = Field(
        default=None,
        description="Iteration termination criteria",
        discriminator="model_type",
    )
    dirimpl: Optional[DIRIMPL] = Field(
        default=None,
        description="Numerical scheme for refraction",
    )
    sigimpl: Optional[SIGIMPL] = Field(
        default=None,
        description="Frequency shifting accuracy",
    )
    ctheta: Optional[CTHETA] = Field(
        default=None,
        description="Prevents excessive directional turning",
    )
    csigma: Optional[CSIGMA] = Field(
        default=None,
        description="Prevents excessive frequency shifting",
    )
    setup: Optional[SETUP] = Field(
        default=None,
        description="Stop criteria in the computation of wave setup",
    )

    def cmd(self) -> str:
        """Command file string for this component."""
        repr = "NUMERIC"
        if self.stop is not None:
            repr += f" {self.stop.render()}"
        if self.dirimpl is not None:
            repr += f" {self.dirimpl.render()}"
        if self.sigimpl is not None:
            repr += f" {self.sigimpl.render()}"
        if self.ctheta is not None:
            repr += f" {self.ctheta.render()}"
        return repr



import os
import shutil
import subprocess
import sys
import tempfile
import zipfile

import pytest
import requests

DATA_REPO = "rom-py/rompy-test-data"
DATA_DIR = os.path.join(os.path.dirname(__file__), "data")
GITHUB_API_RELEASES = f"https://api.github.com/repos/{DATA_REPO}/releases/latest"

# Add the tests directory to the Python path
sys.path.insert(0, os.path.abspath(os.path.dirname(__file__)))

# Import test utilities for logging
from test_utils.logging import configure_test_logging


@pytest.fixture(scope="session", autouse=True)
def docker_available() -> bool:
    try:
        result = subprocess.run(
            [
                "docker",
                "info",
            ],
            capture_output=True,
            text=True,
            timeout=10,
        )
        return result.returncode == 0
    except Exception:
        return False


def pytest_addoption(parser):
    """Add command-line options for pytest."""
    parser.addoption(
        "--run-slow",
        action="store_true",
        default=False,
        help="Run slow tests",
    )
    parser.addoption(
        "--rompy-log-level",
        default="INFO",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        help="Set the logging level for ROMPY tests",
    )


def pytest_configure(config):
    """Configure pytest with plugins and settings, and ensure test data is present."""
    import logging

    # Get log level from command line or use default
    log_level_str = config.getoption("--rompy-log-level")
    getattr(logging, log_level_str)

    # Configure logging for tests
    configure_test_logging(level=log_level_str)


def download_and_extract_data():
    print("Downloading test data from rompy-test-data repo...")
    # Get latest release info
    resp = requests.get(GITHUB_API_RELEASES)
    resp.raise_for_status()
    release = resp.json()
    # Find the zipball URL
    zip_url = release["zipball_url"]
    # Download the zip
    with tempfile.TemporaryDirectory() as tmpdir:
        zip_path = os.path.join(tmpdir, "data.zip")
        with requests.get(zip_url, stream=True) as r:
            r.raise_for_status()
            with open(zip_path, "wb") as f:
                for chunk in r.iter_content(chunk_size=8192):
                    f.write(chunk)
        # Extract only the data/ directory
        with zipfile.ZipFile(zip_path, "r") as zip_ref:
            # Find the top-level directory in the zip
            top_level = zip_ref.namelist()[0].split("/")[0]
            # Extract data/ to DATA_DIR
            for member in zip_ref.namelist():
                if member.startswith(f"{top_level}/data/"):
                    zip_ref.extract(member, tmpdir)
            src_data_dir = os.path.join(tmpdir, top_level, "data")
            if os.path.exists(DATA_DIR):
                shutil.rmtree(DATA_DIR)
            shutil.copytree(src_data_dir, DATA_DIR)
    print("Test data downloaded and extracted.")


# Only download if data dir is missing or empty
if not os.path.exists(DATA_DIR) or not os.listdir(DATA_DIR):
    try:
        download_and_extract_data()
    except Exception as e:
        print(f"Failed to download test data: {e}", file=sys.stderr)
        sys.exit(1)


# Helper to detect CI environment
def is_ci_environment():
    """Detect if running in a CI environment."""
    ci_vars = [
        "CI",
        "GITHUB_ACTIONS",
        "GITLAB_CI",
        "TRAVIS",
        "CIRCLECI",
        "APPVEYOR",
        "BUILDKITE",
        "TEAMCITY_VERSION",
    ]
    return any(
        os.environ.get(var, "").lower() in ("1", "true", "yes", "on") for var in ci_vars
    )


@pytest.fixture(scope="session")
def should_skip_docker_builds():
    """Fixture to determine if Docker builds should be skipped."""
    if os.environ.get("SKIP_DOCKER_BUILDS", "").lower() in ("1", "true", "yes", "on"):
        return True
    if os.environ.get("ROMPY_SKIP_DOCKER_BUILDS", "").lower() in (
        "1",
        "true",
        "yes",
        "on",
    ):
        return True
    if os.environ.get("ROMPY_ENABLE_DOCKER_IN_CI", "").lower() not in (
        "1",
        "true",
        "yes",
        "on",
    ):
        return is_ci_environment()
    return False


@pytest.fixture(scope="session", autouse=True)
def setup_logging():
    """Set up logging for all tests.

    This fixture runs once per test session and ensures that logging is properly
    configured for all tests.
    """
    # Import here to avoid circular imports
    from test_utils.logging import configure_test_logging

    # Configure logging with default settings
    configure_test_logging()

    # Return a function to reconfigure logging if needed
    return configure_test_logging



"""
SWAN Output Subcomponents

This module contains subcomponents for defining output specifications in SWAN,
including spectral output types and frequency representations.
"""

from typing import Literal

from pydantic import Field

from rompy_swan.subcomponents.base import BaseSubComponent


class SPEC1D(BaseSubComponent):
    """Frequency 1D spectra.

    .. code-block:: text

        SPEC1D

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.output import SPEC1D
        spec = SPEC1D()
        print(spec.render())

    """

    model_type: Literal["spec1d", "SPEC1D"] = Field(
        default="spec1d", description="Model type discriminator"
    )

    def cmd(self) -> str:
        """Command file string for this subcomponent."""
        return "SPEC2D"


class SPEC2D(BaseSubComponent):
    """Frequency-direction 2D spectra.

    .. code-block:: text

        SPEC2D

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.output import SPEC2D
        spec = SPEC2D()
        print(spec.render())

    """

    model_type: Literal["spec2d", "SPEC2D"] = Field(
        default="spec2d", description="Model type discriminator"
    )

    def cmd(self) -> str:
        """Command file string for this subcomponent."""
        return "SPEC2D"


class ABS(BaseSubComponent):
    """Absolute frequency spectra.

    .. code-block:: text

        ABS

    Spectra are computed as a function of absolute frequency, i.e., the frequency as
    measured in a fixed point.

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.output import ABS
        freq = ABS()
        print(freq.render())

    """

    model_type: Literal["abs", "ABS"] = Field(
        default="abs", description="Model type discriminator"
    )

    def cmd(self) -> str:
        """Command file string for this subcomponent."""
        return "ABS"


class REL(BaseSubComponent):
    """Relative frequency spectra.

    .. code-block:: text

        REL

    Spectra are computed as a function of relative frequency, i.e., the frequency as
    measured when moving with current.

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.output import REL
        freq = REL()
        print(freq.render())

    """

    model_type: Literal["rel", "REL"] = Field(
        default="rel", description="Model type discriminator"
    )

    def cmd(self) -> str:
        """Command file string for this subcomponent."""
        return "REL"



"""SWAN physics subcomponents."""

from abc import ABC
from typing import Annotated, Literal, Optional

from pydantic import Field, field_validator, model_validator
from pydantic_numpy.typing import Np2DArray

from rompy_swan.subcomponents.base import BaseSubComponent


# ======================================================================================
# Source terms
# ======================================================================================
class SourceTerms(BaseSubComponent, ABC):
    """Source term subcomponent base class."""

    wind_drag: Literal["wu", "fit"] = Field(
        default="wu",
        description="Indicates the wind drag formulation",
    )
    agrow: bool = Field(
        default=False,
        description="Activate the Cavaleri and Malanotte (1981) wave growth term",
    )
    a: Optional[float] = Field(
        default=None,
        description=(
            "Proportionality coefficient when activating the Cavaleri and Malanotte "
            "(1981) wave growth term (SWAN default: 0.0015)"
        ),
    )


class JANSSEN(SourceTerms):
    """Janssen source terms subcomponent.

    .. code-block:: text

        JANSSEN [cds1] [delta] (AGROW [a])

    References
    ----------
    Janssen, P.A., 1989. Wave-induced stress and the drag of air flow over sea waves.
    Journal of Physical Oceanography, 19(6), pp.745-754.

    Janssen, P.A.E.M., Lionello, P. and Zambresky, L., 1989. On the interaction of wind
    and waves. Philosophical transactions of the royal society of London. Series A,
    Mathematical and Physical Sciences, 329(1604), pp.289-301.

    Janssen, P.A., 1991. Quasi-linear theory of wind-wave generation applied to wave
    forecasting. Journal of physical oceanography, 21(11), pp.1631-1642.

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.physics import JANSSEN
        janssen = JANSSEN()
        print(janssen.render())
        janssen = JANSSEN(cds1=4.5, delta=0.5, agrow=True)
        print(janssen.render())

    """

    model_type: Literal["janssen", "JANSSEN"] = Field(
        default="janssen", description="Model type discriminator"
    )
    cds1: Optional[float] = Field(
        default=None,
        description=(
            "Coefficient for determining the rate of whitecapping dissipation "
            "($Cds / s^4_{PM}$) (SWAN default: 4.5)"
        ),
    )
    delta: Optional[float] = Field(
        default=None,
        description=(
            "Coefficient which determines the dependency of the whitecapping on wave "
            "number (mix with Komen et al. formulation) (SWAN default: 0.5)"
        ),
    )

    def cmd(self) -> str:
        repr = "JANSSEN"
        if self.cds1 is not None:
            repr += f" cds1={self.cds1}"
        if self.delta is not None:
            repr += f" delta={self.delta}"
        repr += f" DRAG {self.wind_drag.upper()}"
        if self.agrow:
            repr += " AGROW"
        if self.a is not None and self.agrow:
            repr += f" a={self.a}"
        return repr


class KOMEN(SourceTerms):
    """Komen source terms subcomponent.

    .. code-block:: text

        KOMEN [cds2] [stpm] (AGROW [a])

    References
    ----------
    Komen, G.J., Hasselmann, S. and Hasselmann, K., 1984. On the existence of a fully
    developed wind-sea spectrum. Journal of physical oceanography, 14(8), pp.1271-1285.

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.physics import KOMEN
        komen = KOMEN()
        print(komen.render())
        komen = KOMEN(cds2=2.36e-5, stpm=3.02e-3, agrow=True, a=0.0015)
        print(komen.render())

    """

    model_type: Literal["komen", "KOMEN"] = Field(
        default="komen", description="Model type discriminator"
    )
    cds2: Optional[float] = Field(
        default=None,
        description=(
            "Coefficient for determining the rate of whitecapping dissipation "
            "(`Cds`) (SWAN default: 2.36e-5)"
        ),
    )
    stpm: Optional[float] = Field(
        default=None,
        description=(
            "Value of the wave steepness for a Pierson-Moskowitz spectrum "
            "(`s^2_PM`) (SWAN default: 3.02e-3)"
        ),
    )

    def cmd(self) -> str:
        repr = "KOMEN"
        if self.cds2 is not None:
            repr += f" cds2={self.cds2}"
        if self.stpm is not None:
            repr += f" stpm={self.stpm}"
        repr += f" DRAG {self.wind_drag.upper()}"
        if self.agrow:
            repr += " AGROW"
        if self.a is not None and self.agrow:
            repr += f" a={self.a}"
        return repr


class WESTHUYSEN(SourceTerms):
    """Westhuysen source terms subcomponent.

    .. code-block:: text

        WESTHUYSEN [cds2] [br] (AGROW [a])

    Nonlinear saturation-based whitecapping combined with wind input of Yan (1987).

    Notes
    -----
    The two arguments are specified in the Appendix C of the User manual but not in the
    command description for WESTH in Section 4.5.4. They are also options in the
    WCAPPING command. It is not entirely clear if they should/could be specified here.

    References
    ----------
    van der Westhuysen, A.J., Zijlema, M. and Battjes, J.A., 2007. Nonlinear
    saturation-based whitecapping dissipation in SWAN for deep and shallow water.
    Coastal Engineering, 54(2), pp.151-170.

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.physics import WESTHUYSEN
        westhuysen = WESTHUYSEN()
        print(westhuysen.render())
        westhuysen = WESTHUYSEN(cds2=5.0e-5, br=1.75e-3)
        print(westhuysen.render())

    """

    model_type: Literal["westhuysen", "WESTHUYSEN"] = Field(
        default="westhuysen", description="Model type discriminator"
    )
    cds2: Optional[float] = Field(
        default=None,
        description=(
            "proportionality coefficient due to Alves and Banner (2003) "
            "(SWAN default: 5.0e-5)."
        ),
    )
    br: Optional[float] = Field(
        default=None, description="Threshold saturation level	(SWAN default: 1.75e-3)"
    )

    def cmd(self) -> str:
        repr = "WESTHUYSEN"
        if self.cds2 is not None:
            repr += f" cds2={self.cds2}"
        if self.br is not None:
            repr += f" br={self.br}"
        repr += f" DRAG {self.wind_drag.upper()}"
        if self.agrow:
            repr += " AGROW"
        if self.a is not None and self.agrow:
            repr += f" a={self.a}"
        return repr


class ST6(SourceTerms):
    """St6 source terms subcomponent.

    .. code-block:: text

        ST6 [a1sds] [a2sds] [p1sds] [p2sds] UP|DOWN HWANG|FAN|ECMWF VECTAU|SCATAU &
            TRUE10|U10PROXY [windscaling] DEBIAS [cdfac] (AGROW [a])

    wind input and whitecapping from Rogers et al. (2012) (RBW12).

    Notes
    -----
    The two arguments are specified in the Appendix C of the User manual but not in the
    command description for WESTH in Section 4.5.4. They are also options in the
    WCAPPING command. It is not entirely clear if they should/could be specified here.

    References
    ----------
    Fan, Y., Lin, S.J., Held, I.M., Yu, Z. and Tolman, H.L., 2012. Global ocean surface
    wave simulation using a coupled atmosphere–wave model. Journal of Climate, 25(18),
    pp.6233-6252.

    Rogers, W.E., Babanin, A.V. and Wang, D.W., 2012. Observation-consistent input and
    whitecapping dissipation in a model for wind-generated surface waves: Description
    and simple calculations. Journal of Atmospheric and Oceanic Technology, 29(9),
    pp.1329-1346.

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.physics import ST6
        st6 = ST6(a1sds=4.7e-7, a2sds=6.6e-6)
        print(st6.render())
        kwargs = dict(
            a1sds=2.8e-6,
            a2sds=3.5e-5,
            normalization="up",
            wind_drag="hwang",
            tau="vectau",
            u10="u10proxy",
            windscaling=32.0,
            cdfac=0.89,
            agrow=True,
            a=0.0015,
        )
        st6 = ST6(**kwargs)
        print(st6.render())

    """

    model_type: Literal["st6"] = Field(
        default="st6", description="Model type discriminator"
    )
    a1sds: float = Field(
        description="Coefficient related to local dissipation term T1 (a1 in RBW12)"
    )
    a2sds: float = Field(
        description="Coefficient related to local dissipation term T2 (a2 in RBW12)"
    )
    p1sds: Optional[float] = Field(
        default=None,
        description=(
            "Power coefficient controlling strength of dissipation term T1 "
            "(L in RBW12, SWAN default: 4)"
        ),
    )
    p2sds: Optional[float] = Field(
        default=None,
        description=(
            "Power coefficient controlling strength of dissipation term T2 "
            "(M in RBW12, SWAN default: 4)"
        ),
    )
    normalization: Literal["up", "down"] = Field(
        default="up",
        description=(
            "Selection of normalization of exceedance level by ET(f) (`up`) or E(f) "
            "(`down`) as in RBW12 (right column, page 1333), `up` is default and "
            "strongly recommended"
        ),
    )
    wind_drag: Literal["hwang", "fan", "ecmwf"] = Field(
        default="hwang",
        description=(
            "Wind drag formula, `hwang` is the default and is unchanged from RBW12, "
            "`fan` is from Fan et al. (2012), `ecmwf` follows WAM Cycle 4 methodology"
        ),
    )
    tau: Literal["vectau", "scatau"] = Field(
        default="vectau",
        description=(
            "Use vector (vectau) or scalar (scatau) calculation for the wind strerss "
            "(Eq. 12 in RBW12), `vectau` is the default and strongly recommended"
        ),
    )
    u10: Literal["u10proxy", "true10"] = Field(
        default="u10proxy",
        description="Wind velocity definition",
    )
    windscaling: Optional[float] = Field(
        default=32.0,
        description="Factor to scale U10 with U* when using U10PROXY",
    )
    cdfac: Optional[float] = Field(
        default=None,
        description=(
            "Counter bias in the input wind fields by providing a multiplier "
            "on the drag coefficient"
        ),
        gt=0.0,
    )

    @model_validator(mode="after")
    def debias_only_with_hwang(self) -> "ST6":
        if self.cdfac is not None and self.wind_drag != "hwang":
            raise ValueError(
                f"Debias is only supported with hwang wind drag, not {self.wind_drag}"
            )
        return self

    @property
    def u10_cmd(self) -> str:
        if self.u10 == "true10":
            return "TRUE10"
        else:
            return f"U10PROXY windscaling={self.windscaling}"

    def cmd(self) -> str:
        repr = f"ST6 a1sds={self.a1sds} a2sds={self.a2sds}"
        if self.p1sds is not None:
            repr += f" p1sds={self.p1sds}"
        if self.p2sds is not None:
            repr += f" p2sds={self.p2sds}"
        repr += f" {self.normalization.upper()}"
        repr += f" {self.wind_drag.upper()}"
        repr += f" {self.tau.upper()}"
        repr += f" {self.u10_cmd}"
        if self.cdfac is not None:
            repr += f" DEBIAS cdfac={self.cdfac}"
        if self.agrow:
            repr += " AGROW"
        if self.a is not None and self.agrow:
            repr += f" a={self.a}"
        return repr


class ST6C1(ST6):
    """First ST6 calibration in the SWAN user manual.

    .. code-block:: text

        ST6 4.7e-7 6.6e-6 4.0 4.0 UP HWANG VECTAU U10PROXY 28.0 AGROW

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.physics import ST6C1
        st6 = ST6C1()
        print(st6.render())

    """

    model_type: Literal["st6c1"] = Field(default="st6c1")
    a1sds: Literal[4.7e-7] = Field(default=4.7e-7)
    a2sds: Literal[6.6e-6] = Field(default=6.6e-6)
    p1sds: Literal[4.0] = Field(default=4.0)
    p2sds: Literal[4.0] = Field(default=4.0)
    windscaling: Literal[28.0] = Field(default=28.0)
    agrow: Literal[True] = Field(default=True)


class ST6C2(ST6C1):
    """Second ST6 calibration in the SWAN user manual.

    .. code-block:: text

        ST6 4.7e-7 6.6e-6 4.0 4.0 UP FAN VECTAU U10PROXY 28.0 AGROW

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.physics import ST6C2
        st6 = ST6C2()
        print(st6.render())

    TODO: Ensure validator is reused here so fan and debias are not used together.

    """

    model_type: Literal["st6c2"] = Field(default="st6c2")
    wind_drag: Literal["fan"] = Field(default="fan")


class ST6C3(ST6C1):
    """Third ST6 calibration in the SWAN user manual.

    .. code-block:: text

        ST6 2.8e-6 3.5e-5 4.0 4.0 UP HWANG VECTAU U10PROXY 32.0 AGROW

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.physics import ST6C3
        st6 = ST6C3()
        print(st6.render())

    """

    model_type: Literal["st6c3"] = Field(default="st6c3")
    a1sds: Literal[2.8e-6] = Field(default=2.8e-6)
    a2sds: Literal[3.5e-5] = Field(default=3.5e-5)
    windscaling: Literal[32.0] = Field(default=32.0)


class ST6C4(ST6C3):
    """Fourth ST6 calibration in the SWAN user manual.

    .. code-block:: text

        ST6 2.8e-6 3.5e-5 4.0 4.0 UP HWANG VECTAU U10PROXY 32.0 DEBIAS 0.89 AGROW

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.physics import ST6C4
        st6 = ST6C4()
        print(st6.render())

    """

    model_type: Literal["st6c4"] = Field(default="st6c4")
    cdfac: Literal[0.89] = Field(default=0.89)


class ST6C5(ST6C1):
    """Fifth ST6 calibration in the SWAN user manual.

    .. code-block:: text

        ST6 4.7e-7 6.6e-6 4.0 4.0 UP HWANG VECTAU U10PROXY 28.0 AGROW

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.physics import ST6C5
        st6 = ST6C5()
        print(st6.render())

    """

    model_type: Literal["st6c5"] = Field(default="st6c5")
    cdfac: Literal[0.89] = Field(default=0.89)
    a1sds: Literal[6.5e-6] = Field(default=6.5e-6)
    a2sds: Literal[8.5e-5] = Field(default=8.5e-5)
    windscaling: Literal[35.0] = Field(default=35.0)


# =====================================================================================
# Biphase
# =====================================================================================
class ELDEBERKY(BaseSubComponent):
    """Biphase of Eldeberky (1999).

    .. code-block:: text

        BIPHASE ELDEBERKY [urcrit]

    Biphase parameterisation as a funtion of the Ursell number of Eldeberky (1999).

    References
    ----------
    Eldeberky, Y., Polnikov, V. and Battjes, J.A., 1996. A statistical approach for
    modeling triad interactions in dispersive waves. In Coastal Engineering 1996
    (pp. 1088-1101).

    Eldeberky, Y. and Madsen, P.A., 1999. Deterministic and stochastic evolution
    equations for fully dispersive and weakly nonlinear waves. Coastal Engineering,
    38(1), pp.1-24.

    Doering, J.C. and Bowen, A.J., 1995. Parametrization of orbital velocity
    asymmetries of shoaling and breaking waves using bispectral analysis. Coastal
    engineering, 26(1-2), pp.15-33.

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.physics import ELDEBERKY
        biphase = ELDEBERKY()
        print(biphase.render())
        biphase = ELDEBERKY(urcrit=0.63)
        print(biphase.render())

    """

    model_type: Literal["eldeberky"] = Field(
        default="eldeberky", description="Model type discriminator"
    )
    urcrit: Optional[float] = Field(
        default=None,
        description=(
            "The critical Ursell number appearing in the parametrization. Note: the "
            "value of `urcrit` is setted by Eldeberky (1996) at 0.2 based on a "
            "laboratory experiment, whereas Doering and Bowen (1995) employed the "
            "value of 0.63 based on the field experiment data (SWAN default: 0.63)"
        ),
    )

    def cmd(self) -> str:
        repr = "BIPHASE ELDEBERKY"
        if self.urcrit is not None:
            repr += f" urcrit={self.urcrit}"
        return repr


class DEWIT(BaseSubComponent):
    """Biphase of De Wit (2022).

    .. code-block:: text

        BIPHASE DEWIT [lpar]

    Biphase parameterization based on bed slope and peak period of De Wit (2022).

    References
    ----------
    De Wit, F.P., 2022. Wave shape prediction in complex coastal systems (Doctoral
    dissertation, PhD. thesis. Delft University of Technology. https://repository.
    tudelft. nl/islandora/object/uuid% 3A0fb850a4-4294-4181-9d74-857de21265c2).

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.physics import DEWIT
        biphase = DEWIT()
        print(biphase.render())
        biphase = DEWIT(lpar=0.0)
        print(biphase.render())

    """

    model_type: Literal["dewit"] = Field(
        default="dewit", description="Model type discriminator"
    )
    lpar: Optional[float] = Field(
        default=None,
        description=(
            "Scales spatial averaging of the De Wit's biphase in terms of a multiple "
            "of peak wave length of the incident wave field. Note: `lpar` = 0` means "
            "no averaging (SWAN default: 0)"
        ),
    )

    def cmd(self) -> str:
        repr = "BIPHASE DEWIT"
        if self.lpar is not None:
            repr += f" lpar={self.lpar}"
        return repr


# =====================================================================================
# Transmission and reflection
# =====================================================================================
class TRANSM(BaseSubComponent):
    """Constant transmission coefficient.

    .. code-block:: text

        TRANSM [trcoef]

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.physics import TRANSM
        transm = TRANSM()
        print(transm.render())
        transm = TRANSM(trcoef=0.5)
        print(transm.render())

    """

    model_type: Literal["transm", "TRANSM"] = Field(
        default="transm", description="Model type discriminator"
    )
    trcoef: Optional[float] = Field(
        default=None,
        description=(
            "Constant transmission coefficient (ratio of transmitted over incoming "
            "significant wave height) (SWAN default: 0.0) (no transmission = complete "
            "blockage)"
        ),
        ge=0.0,
        le=1.0,
    )

    def cmd(self) -> str:
        """Command file string for this subcomponent."""
        repr = "TRANSM"
        if self.trcoef is not None:
            repr += f" trcoef={self.trcoef}"
        return repr


class TRANS1D(BaseSubComponent):
    """Frequency dependent transmission.

    .. code-block:: text

        TRANS1D < [trcoef] >

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.physics import TRANS1D
        transm = TRANS1D(trcoef=[0.0, 0.0, 0.2, 0.5, 0.2, 0.0, 0.0])
        print(transm.render())

    """

    model_type: Literal["trans1d", "TRANS1D"] = Field(
        default="trans1d", description="Model type discriminator"
    )
    trcoef: list[Annotated[float, Field(ge=0.0, le=1.0)]] = Field(
        description=(
            "Transmission coefficient (ratio of transmitted over incoming significant "
            "wave height) per frequency. The number of these transmission values must "
            "be equal to the number of frequencies, i.e. `msc` + 1"
        ),
    )

    def cmd(self) -> str:
        """Command file string for this subcomponent."""
        return f"TRANS1D {' '.join(str(v) for v in self.trcoef)}"


class TRANS2D(BaseSubComponent):
    """Frequency-direction dependent transmission.

    .. code-block:: text

        TRANS2D < [trcoef] >

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.physics import TRANS2D
        trcoef = np.array([[0.0, 0.0], [0.1, 0.1], [0.2, 0.2]])
        transm = TRANS2D(trcoef=trcoef)
        print(transm.render())

    """

    model_type: Literal["trans2d", "TRANS2D"] = Field(
        default="trans2d", description="Model type discriminator"
    )
    trcoef: Np2DArray = Field(
        description=(
            "Transmission coefficient (ratio of transmitted over incoming significant "
            "wave height) per frequency and direction, rows represent directions and "
            "columns represent frequencies"
        ),
    )

    @field_validator("trcoef")
    @classmethod
    def constrained_0_1(cls, value: float) -> float:
        """Ensure all directions have the same number of frequencies."""
        if value.min() < 0 or value.max() > 1:
            raise ValueError("Transmission coefficients must be between 0.0 and 1.0")
        return value

    def cmd(self) -> str:
        """Command file string for this subcomponent."""
        repr = "TRANS2D"
        for coef in self.trcoef:
            repr += f" &\n\t{' '.join(str(v) for v in coef)}"
        return f"{repr} &\n\t"


class GODA(BaseSubComponent):
    """DAM transmission of Goda/Seelig (1979).

    .. code-block:: text

        DAM GODA [hgt] [alpha] [beta]

    This option specified transmission coefficients dependent on the incident wave
    conditions at the obstacle and on the obstacle height (which may be submerged).

    References
    ----------
    Goda, Y. and Suzuki, Y., 1976. Estimation of incident and reflected waves in random
    wave experiments. In Coastal Engineering 1976 (pp. 828-845).

    Seelig, W.N., 1979. Effects of breakwaters on waves: Laboratory test of wave
    transmission by overtopping. In Proc. Conf. Coastal Structures, 1979
    (Vol. 79, No. 2, pp. 941-961).

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.physics import GODA
        transm = GODA(hgt=3.0)
        print(transm.render())
        transm = GODA(hgt=3.0, alpha=2.6, beta=0.15)
        print(transm.render())

    """

    model_type: Literal["goda", "GODA"] = Field(
        default="goda", description="Model type discriminator"
    )
    hgt: float = Field(
        description=(
            "The elevation of the top of the obstacle above reference level (same "
            "reference level as for bottom etc.); use a negative value if the top is "
            "below that reference level"
        ),
    )
    alpha: Optional[float] = Field(
        default=None,
        description=(
            "coefficient determining the transmission coefficient for Goda's "
            "transmission formula (SWAN default: 2.6)"
        ),
    )
    beta: Optional[float] = Field(
        default=None,
        description=(
            "Another coefficient determining the transmission coefficient for Goda's "
            "transmission formula (SWAN default: 0.15)"
        ),
    )

    def cmd(self) -> str:
        """Command file string for this subcomponent."""
        repr = f"DAM {self.model_type.upper()} hgt={self.hgt}"
        if self.alpha is not None:
            repr += f" alpha={self.alpha}"
        if self.beta is not None:
            repr += f" beta={self.beta}"
        return repr


class DANGREMOND(BaseSubComponent):
    """DAM transmission of d'Angremond et al. (1996).

    .. code-block:: text

        DAM DANGREMOND [hgt] [slope] [Bk]

    This option specifies transmission coefficients dependent on the incident wave
    conditions at the obstacle and on the obstacle height (which may be submerged).

    References
    ----------
    d'Angremond, K., Van Der Meer, J.W. and De Jong, R.J., 1996. Wave transmission at
    low-crested structures. In Coastal Engineering 1996 (pp. 2418-2427).

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.physics import DANGREMOND
        transm = DANGREMOND(hgt=3.0, slope=60, Bk=10.0)
        print(transm.render())

    """

    model_type: Literal["dangremond", "DANGREMOND"] = Field(
        default="dangremond", description="Model type discriminator"
    )
    hgt: float = Field(
        description=(
            "The elevation of the top of the obstacle above reference level (same "
            "reference level as for bottom etc.); use a negative value if the top is "
            "below that reference level"
        ),
    )
    slope: float = Field(
        description="The slope of the obstacle (in degrees)", ge=0.0, le=90.0
    )
    Bk: float = Field(description="The crest width of the obstacle")

    def cmd(self) -> str:
        """Command file string for this subcomponent."""
        repr = f"DAM {self.model_type.upper()}"
        repr += f" hgt={self.hgt} slope={self.slope} Bk={self.Bk}"
        return repr


class REFL(BaseSubComponent):
    """Obstacle reflections.

    .. code-block:: text

        REFL [reflc]

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.physics import REFL
        refl = REFL()
        print(refl.render())
        refl = REFL(reflc=0.5)
        print(refl.render())

    """

    model_type: Literal["refl", "REFL"] = Field(
        default="refl", description="Model type discriminator"
    )
    reflc: Optional[float] = Field(
        default=None,
        description=(
            "Constant reflection coefficient (ratio of reflected over incoming "
            "significant wave height) (SWAN default: 1.0)"
        ),
    )

    def cmd(self) -> str:
        """Command file string for this subcomponent."""
        repr = "REFL"
        if self.reflc is not None:
            repr += f" reflc={self.reflc}"
        return repr


class RSPEC(BaseSubComponent):
    """Specular reflection.

    .. code-block:: text

        RSPEC

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.physics import RSPEC
        refl = RSPEC()
        print(refl.render())

    """

    model_type: Literal["rspec", "RSPEC"] = Field(
        default="rspec", description="Model type discriminator"
    )

    def cmd(self) -> str:
        """Command file string for this subcomponent."""
        return "RSPEC"


class RDIFF(BaseSubComponent):
    """Diffuse reflection.

    .. code-block:: text

        RDIFF [pown]

    Specular reflection where incident waves are scattered over reflected direction.

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.physics import RDIFF
        refl = RDIFF()
        print(refl.render())
        refl = RDIFF(pown=1.0)
        print(refl.render())

    """

    model_type: Literal["rdiff", "RDIFF"] = Field(
        default="rdiff", description="Model type discriminator"
    )
    pown: Optional[float] = Field(
        default=None,
        description=(
            "Each incoming direction θ is scattered over reflected direction θ_refl "
            "according to cos^pown(θ-θ_refl). The parameter `pown` indicates the width"
            "of the redistribution function (SWAN default: 1.0)"
        ),
    )

    def cmd(self) -> str:
        """Command file string for this subcomponent."""
        repr = "RDIFF"
        if self.pown is not None:
            repr += f" pown={self.pown}"
        return repr


class FREEBOARD(BaseSubComponent):
    """Freeboard dependent transmission and reflection.

    .. code-block:: text

        FREEBOARD [hgt] [gammat] [gammar] [QUAY]

    With this option the user indicates that the fixed transmission `trcoef` and
    reflection `reflc` coefficients are freeboard dependent. The freeboard dependency
    has no effect on the transmission coefficient as computed using the DAM option.

    Notes
    -----
    See the Scientific/Technical documentation for background information on the
    `gammat` and `gammar` shape parameters.

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.physics import FREEBOARD
        freeboard = FREEBOARD(hgt=2.0)
        print(freeboard.render())
        freeboard = FREEBOARD(hgt=2.0, gammat=1.0, gammar=1.0, quay=True)
        print(freeboard.render())

    """

    model_type: Literal["freeboard", "FREEBOARD"] = Field(
        default="freeboard", description="Model type discriminator"
    )
    hgt: float = Field(
        description=(
            "The elevation of the top of the obstacle or height of the quay above the "
            "reference level (same reference level as for the bottom). Use a negative "
            "value if the top is below that reference level. In case `hgt` is also "
            "specified in the DAM option, both values of `hgt` should be equal for "
            "consistency"
        ),
    )
    gammat: Optional[float] = Field(
        default=None,
        description=(
            "Shape parameter of relative freeboard dependency of transmission "
            "coefficient. This parameter should be higher than zero (SWAN default 1.0)"
        ),
        gt=0.0,
    )
    gammar: Optional[float] = Field(
        default=None,
        description=(
            "Shape parameter of relative freeboard dependency of reflection "
            "coefficient. This parameter should be higher than zero (SWAN default 1.0)"
        ),
        gt=0.0,
    )
    quay: bool = Field(
        default=False,
        description=(
            "With this option the user indicates that the freeboard dependency of the "
            "transmission and reflection coefficients also depends on the relative "
            "position of an obstacle-linked grid point with respect to the position "
            "of the obstacle line representing the edge of a quay. In case the active "
            "grid point is on the deeper side of the obstacle, then the correction "
            "factors are applied using the parameters `hgt`, `gammat` and `gammar`."
            "In case the active grid point is on the shallower side of the obstacle, "
            "the reflection coefficient is set to 0 and the transmission coefficient "
            "to 1."
        ),
    )

    def cmd(self) -> str:
        """Command file string for this subcomponent."""
        repr = "FREEBOARD"
        if self.hgt is not None:
            repr += f" hgt={self.hgt}"
        if self.gammat is not None:
            repr += f" gammat={self.gammat}"
        if self.gammar is not None:
            repr += f" gammar={self.gammar}"
        if self.quay:
            repr += " QUAY"
        return repr


class LINE(BaseSubComponent):
    """Line of points to define obstacle location.

    .. code-block:: text

        LINE < [xp] [yp] >

    With this option the user indicates that the fixed transmission `trcoef` and
    reflection `reflc` coefficients are freeboard dependent. The freeboard dependency
    has no effect on the transmission coefficient as computed using the DAM option.

    Notes
    -----
    Points coordinates should be provided in m If Cartesian coordinates are used or in
    degrees if spherical coordinates are used (see command `COORD`). At least two
    corner points must be provided.

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.subcomponents.physics import LINE
        line = LINE(xp=[174.1, 174.2, 174.3], yp=[-39.1, -39.1, -39.1])
        print(line.render())

    """

    model_type: Literal["line", "LINE"] = Field(
        default="line", description="Model type discriminator"
    )
    xp: list[float] = Field(
        description="The x-coordinates of the points defining the line", min_length=2
    )
    yp: list[float] = Field(
        description="The y-coordinates of the points defining the line", min_length=2
    )

    @model_validator(mode="after")
    def check_length(self) -> "LINE":
        """Check that the length of xp and yp are the same."""
        if len(self.xp) != len(self.yp):
            raise ValueError("xp and yp must be the same length")
        return self

    def cmd(self) -> str:
        """Command file string for this subcomponent."""
        repr = "LINE"
        for xp, yp in zip(self.xp, self.yp):
            repr += f" {xp} {yp}"
        return repr



"""Test physics sub-components."""

import pytest
from pydantic import ValidationError

# Import test utilities
from test_utils.logging import get_test_logger

# Initialize logger
logger = get_test_logger(__name__)
from rompy_swan.subcomponents.physics import (
    DANGREMOND,
    DEWIT,
    ELDEBERKY,
    FREEBOARD,
    GODA,
    JANSSEN,
    KOMEN,
    LINE,
    RDIFF,
    REFL,
    RSPEC,
    ST6,
    ST6C1,
    ST6C2,
    ST6C3,
    ST6C4,
    ST6C5,
    TRANS1D,
    TRANS2D,
    TRANSM,
    WESTHUYSEN,
)


# =====================================================================================
# Source terms
# =====================================================================================
def test_janssen():
    jans = JANSSEN()
    assert jans.render() == "JANSSEN DRAG WU"
    jans = JANSSEN(cds1=4.5, delta=0.5, wind_drag="fit", agrow=True, a=0.0015)
    assert jans.render() == "JANSSEN cds1=4.5 delta=0.5 DRAG FIT AGROW a=0.0015"


def test_komen():
    kom = KOMEN()
    assert kom.render() == "KOMEN DRAG WU"
    kom = KOMEN(cds2=2.36e-5, stpm=3.02e-3, wind_drag="fit", agrow=True, a=0.0020)
    assert kom.render() == "KOMEN cds2=2.36e-05 stpm=0.00302 DRAG FIT AGROW a=0.002"


def test_westhuysen():
    west = WESTHUYSEN()
    assert west.render() == "WESTHUYSEN DRAG WU"
    west = WESTHUYSEN(cds2=5.0e-5, br=1.7e-3, wind_drag="fit", agrow=True, a=0.0015)
    assert west.render() == "WESTHUYSEN cds2=5e-05 br=0.0017 DRAG FIT AGROW a=0.0015"


def test_st6_default():
    st6 = ST6(
        a1sds=4.7e-7,
        a2sds=6.6e-6,
    )
    assert st6.render() == (
        "ST6 a1sds=4.7e-07 a2sds=6.6e-06 UP HWANG VECTAU U10PROXY windscaling=32.0"
    )


def test_st6_c1():
    st6 = ST6C1()
    assert st6.render() == (
        "ST6 a1sds=4.7e-07 a2sds=6.6e-06 p1sds=4.0 p2sds=4.0 UP HWANG VECTAU U10PROXY "
        "windscaling=28.0 AGROW"
    )


def test_st6_c2():
    st6 = ST6C2()
    assert st6.render() == (
        "ST6 a1sds=4.7e-07 a2sds=6.6e-06 p1sds=4.0 p2sds=4.0 UP FAN VECTAU U10PROXY "
        "windscaling=28.0 AGROW"
    )


def test_st6_c3():
    st6 = ST6C3()
    assert st6.render() == (
        "ST6 a1sds=2.8e-06 a2sds=3.5e-05 p1sds=4.0 p2sds=4.0 UP HWANG VECTAU U10PROXY "
        "windscaling=32.0 AGROW"
    )


def test_st6_c4():
    st6 = ST6C4()
    assert st6.render() == (
        "ST6 a1sds=2.8e-06 a2sds=3.5e-05 p1sds=4.0 p2sds=4.0 UP HWANG VECTAU U10PROXY "
        "windscaling=32.0 DEBIAS cdfac=0.89 AGROW"
    )


def test_st6_c5():
    st6 = ST6C5()
    assert st6.render() == (
        "ST6 a1sds=6.5e-06 a2sds=8.5e-05 p1sds=4.0 p2sds=4.0 UP HWANG VECTAU U10PROXY "
        "windscaling=35.0 DEBIAS cdfac=0.89 AGROW"
    )


# =====================================================================================
# Triads
# =====================================================================================
def test_biphase_elderberky():
    biphase = ELDEBERKY()
    assert biphase.render() == "BIPHASE ELDEBERKY"
    biphase = ELDEBERKY(urcrit=0.63)
    assert biphase.render() == "BIPHASE ELDEBERKY urcrit=0.63"


def test_biphase_dewit():
    biphase = DEWIT()
    assert biphase.render() == "BIPHASE DEWIT"
    biphase = DEWIT(lpar=0)
    assert biphase.render() == "BIPHASE DEWIT lpar=0.0"


# =====================================================================================
# Obstacle
# =====================================================================================
def test_transm():
    trans = TRANSM()
    assert trans.render() == "TRANSM"
    trans = TRANSM(trcoef=0.0)
    assert trans.render() == "TRANSM trcoef=0.0"
    with pytest.raises(ValidationError):
        TRANSM(trcoef=1.1)


def test_trans1d():
    trans = TRANS1D(trcoef=[0.0, 0.0, 0.3, 0.2])
    assert trans.render() == "TRANS1D 0.0 0.0 0.3 0.2"
    with pytest.raises(ValidationError):
        TRANS1D(trcoef=[1.1, 0.0, 0.3, 0.2, 0.1])


def test_trans2d():
    trans = TRANS2D(trcoef=[[0.0, 0.0, 0.0], [0.1, 0.1, 0.1]])
    assert "TRANS2D" in trans.render()
    "0.0 0.0 0.0" in trans.render()
    "0.1 0.1 0.1" in trans.render()
    with pytest.raises(ValidationError):
        TRANS2D(trcoef=[[0.0, 0.0, 0.0], [0.1, 0.1, 0.1, 0.1]])
    with pytest.raises(ValidationError):
        TRANS2D(trcoef=[[0.0, 0.0, 0.0], [0.1, 0.1, 1.1]])


def test_goda():
    trans = GODA(hgt=1.0)
    assert trans.render() == "DAM GODA hgt=1.0"
    trans = GODA(hgt=-1.0, alpha=2.6, beta=0.15)
    assert trans.render() == "DAM GODA hgt=-1.0 alpha=2.6 beta=0.15"


def test_dangremond():
    trans = DANGREMOND(hgt=3.0, slope=60, Bk=10.0)
    assert trans.render() == "DAM DANGREMOND hgt=3.0 slope=60.0 Bk=10.0"
    with pytest.raises(ValidationError):
        DANGREMOND(hgt=3.0, slope=120, Bk=1.0)


def test_refl():
    refl = REFL()
    assert refl.render() == "REFL"
    refl = REFL(reflc=0.5)
    assert refl.render() == "REFL reflc=0.5"


def test_rspec():
    refl = RSPEC()
    assert refl.render() == "RSPEC"


def test_rdiff():
    refl = RDIFF()
    assert refl.render() == "RDIFF"
    refl = RDIFF(pown=1.0)
    assert refl.render() == "RDIFF pown=1.0"


def test_freeboard():
    free = FREEBOARD(
        hgt=2.0,
        gammat=0.5,
        gammar=0.5,
        quay=True,
    )
    assert free.render() == "FREEBOARD hgt=2.0 gammat=0.5 gammar=0.5 QUAY"


def test_freeboard_no_quay():
    free = FREEBOARD(
        hgt=2.0,
        gammat=0.5,
        gammar=0.5,
        quay=False,
    )
    assert free.render() == "FREEBOARD hgt=2.0 gammat=0.5 gammar=0.5"


def test_freeboard_gamma_gt_0():
    with pytest.raises(ValidationError):
        FREEBOARD(hgt=2.0, gammat=0.0)
    with pytest.raises(ValidationError):
        FREEBOARD(hgt=2.0, gammar=0.0)


def test_line():
    line = LINE(xp=[174.1, 174.2, 174.3], yp=[-39.1, -39.1, -39.1])
    assert [f"{x} {y}" in line.render() for x, y in zip(line.xp, line.yp)]



from pathlib import Path

import pytest
from rompy.backends.config import DockerConfig
from rompy.model import ModelRun
from rompy.run.docker import DockerRunBackend


@pytest.mark.slow
def test_swan_container_basic_config(
    tmp_path, docker_available, should_skip_docker_builds
):
    if not docker_available:
        pytest.skip("Docker not available")
    if should_skip_docker_builds:
        pytest.skip("Skipping Potential Docker build tests in CI environment")
    """Test SWAN container with framework integration - validates template rendering and Docker execution."""
    from rompy_swan.components.boundary import BOUNDSPEC, INITIAL
    from rompy_swan.components.cgrid import REGULAR
    from rompy_swan.components.group import INPGRIDS, LOCKUP, OUTPUT
    from rompy_swan.components.inpgrid import CURVILINEAR, INPGRID, READINP
    from rompy_swan.components.inpgrid import REGULAR as INPGRID_REGULAR
    from rompy_swan.components.inpgrid import UNSTRUCTURED
    from rompy_swan.components.lockup import COMPUTE_NONSTAT
    from rompy_swan.components.output import BLOCK
    from rompy_swan.components.physics import (BREAKING_CONSTANT,
                                               FRICTION_MADSEN, GEN3)
    from rompy_swan.components.startup import COORDINATES, MODE, PROJECT, SET
    from rompy_swan.config import SwanConfigComponents
    from rompy_swan.subcomponents.boundary import CONSTANTPAR, DEFAULT, SIDE
    from rompy_swan.subcomponents.physics import ST6
    from rompy_swan.subcomponents.readgrid import GRIDREGULAR, READINP
    from rompy_swan.subcomponents.spectrum import PM, SHAPESPEC, SPECTRUM
    from rompy_swan.subcomponents.startup import SPHERICAL
    from rompy_swan.subcomponents.time import NONSTATIONARY

    # Create cgrid component using actual component classes
    cgrid_config = REGULAR(
        spectrum=SPECTRUM(mdc=36, flow=0.04, fhigh=0.4),
        grid=GRIDREGULAR(
            xp=115.68,  # Grid origin x
            yp=-32.76,  # Grid origin y
            alp=0.0,  # Grid rotation
            xlen=0.05,  # Grid length x (50 * 0.001)
            ylen=0.03,  # Grid length y (30 * 0.001)
            mx=50,  # Number of grid points x
            my=30,  # Number of grid points y
        ),
    )

    # Synthetic bathymetry and wind using inpgrids approach with actual component classes
    inpgrid_config = INPGRIDS(
        inpgrids=[
            INPGRID_REGULAR(
                grid_type="bottom",
                xpinp=115.68,
                ypinp=-32.76,
                alpinp=0.0,
                mxinp=50,
                myinp=30,
                dxinp=0.001,
                dyinp=0.001,
                excval=-999.0,
                readinp=READINP(grid_type="bottom", fname1="bottom.txt"),
            ),
            INPGRID_REGULAR(
                grid_type="wind",
                xpinp=115.68,
                ypinp=-32.76,
                alpinp=0.0,
                mxinp=50,
                myinp=30,
                dxinp=0.001,
                dyinp=0.001,
                excval=-999.0,
                readinp=READINP(grid_type="wind", fname1="wind.txt"),
                nonstationary=NONSTATIONARY(
                    tbeg="2023-01-01T00:00:00",
                    tend="2023-01-01T06:00:00",
                    delt="PT1H",
                    tfmt=1,
                    dfmt="hr",
                ),
            ),
        ]
    )

    # Simple boundary forcing configuration using actual component classes
    boundary_config = BOUNDSPEC(
        shapespec=SHAPESPEC(shape=PM()),  # PM spectrum
        location=SIDE(side="west", direction="ccw"),
        data=CONSTANTPAR(
            hs=1.0,  # Significant wave height 1m
            per=8.0,  # Wave period 8s
            dir=90.0,  # Wave direction from east (90 degrees)
            dd=15.0,  # Directional spread 15 degrees
        ),
    )

    # Basic startup configuration using actual component classes
    startup_config = {
        "project": PROJECT(name="Test Container", nr="0001"),  # Max 16 chars
        "set": SET(
            level=0.0,
            direction_convention="nautical",
            maxerr=3,  # Maximum allowed error tolerance
        ),
        "mode": MODE(kind="nonstationary", dim="twodimensional"),
        "coordinates": COORDINATES(kind=SPHERICAL()),
    }

    # Physics configuration using actual component classes
    physics_config = {
        "gen": GEN3(
            source_terms=ST6(
                a1sds=4.75e-7,  # Required ST6 parameter
                a2sds=7.0e-5,  # Required ST6 parameter
            )
        ),
        "friction": FRICTION_MADSEN(kn=0.015),
        "breaking": BREAKING_CONSTANT(alpha=1.0, gamma=0.73),
    }

    # Initial condition (cold start) using actual component classes
    initial_config = INITIAL(kind=DEFAULT())

    # Simple output configuration using actual component classes
    output_config = OUTPUT(
        block=BLOCK(
            sname="COMPGRID",
            fname="./swangrid.nc",
            output=["depth", "hsign", "tps", "dir"],
            times={"dfmt": "min"},
        )
    )

    # Lockup configuration with COMPUTE command using actual component classes
    lockup_config = LOCKUP(
        compute=COMPUTE_NONSTAT(
            times=NONSTATIONARY(
                tbeg="2023-01-01T00:00:00",
                tend="2023-01-01T06:00:00",
                delt="PT1H",
                tfmt=1,
                dfmt="min",
            )
        )
    )

    # Simple synthetic bathymetry and wind file generation
    def create_synthetic_files(staging_dir):
        """Create synthetic bathymetry and wind files for testing."""
        # Create bathymetry file with uniform 10m depth
        bottom_file = staging_dir / "bottom.txt"
        depths = []
        for j in range(30):  # my points
            for i in range(50):  # mx points
                depths.append("10.0")
        bottom_file.write_text("\n".join(depths) + "\n")

        # Create wind file with constant 10 m/s easterly wind (u=10, v=0)
        wind_file = staging_dir / "wind.txt"
        winds = []
        for j in range(30):  # my points
            for i in range(50):  # mx points
                winds.append("10.0 0.0")  # u=10 m/s (easterly), v=0 m/s
        wind_file.write_text("\n".join(winds) + "\n")

        return bottom_file, wind_file

    config = SwanConfigComponents(
        startup=startup_config,
        cgrid=cgrid_config,
        inpgrid=inpgrid_config,
        boundary=boundary_config,
        initial=initial_config,
        physics=physics_config,
        output=output_config,
        lockup=lockup_config,
    )

    model_run = ModelRun(
        run_id="test_swan_container",
        period=dict(start="20230101T00", duration="6h", interval="1h"),
        output_dir=str(tmp_path),
        config=config,
    )

    # Generate model files
    model_run.generate()

    # Create synthetic bathymetry and wind files after generation
    staging_dir = Path(model_run.staging_dir)
    bottom_file, wind_file = create_synthetic_files(staging_dir)
    print(f"Created synthetic files: {bottom_file.name}, {wind_file.name}")

    # Use single processor to avoid MPI root issues and segfaults
    run_cmd = 'bash -c "cd /app/run_id && swan.exe"'

    # Get dockerfile paths for DockerRunBackend to handle building if needed
    repo_root = Path(__file__).resolve().parents[1]
    context_path = repo_root / "docker"

    docker_config = DockerConfig(
        dockerfile=Path("Dockerfile"),  # Relative to build context
        build_context=context_path,
        executable=run_cmd,
        cpu=1,  # Single CPU since we're not using MPI
    )

    result = model_run.run(backend=docker_config)

    # Note: result may be False due to SWAN segfault (no wave forcing),
    # but this test validates framework integration and template fixes

    generated_dir = Path(model_run.output_dir) / model_run.run_id

    # Main success: INPUT file generated properly using SwanConfigComponents
    input_file = generated_dir / "INPUT"
    assert input_file.exists(), "INPUT file should exist"

    input_content = input_file.read_text()
    assert "COMPUTE NONST" in input_content, "COMPUTE command should be present"

    # Check if SWAN produced any output files (bonus if it works)
    output_files = list((generated_dir).glob("*.nc"))
    assert output_files, "No SWAN output .nc files found in generated directory"

    # Verify output file structure using xarray (similar to SCHISM test)
    import numpy as np
    import xarray as xr

    # Check the main output file
    main_output = output_files[0]  # Usually swangrid.nc
    ds = xr.open_dataset(main_output)
    print(ds)

    # Check for required dimensions
    assert "time" in ds.dims, "Missing 'time' dimension in SWAN output"
    assert "longitude" in ds.dims, "Missing 'longitude' dimension in SWAN output"
    assert "latitude" in ds.dims, "Missing 'latitude' dimension in SWAN output"

    # Check for key wave variables
    assert (
        "hs" in ds.data_vars
    ), "Missing significant wave height 'hs' variable in SWAN output"
    assert "depth" in ds.data_vars, "Missing 'depth' variable in SWAN output"

    # Check dimensions are reasonable
    assert ds.dims["time"] > 0, "Time dimension should be positive"
    assert ds.dims["longitude"] > 0, "Longitude dimension should be positive"
    assert ds.dims["latitude"] > 0, "Latitude dimension should be positive"

    # Check that we have some non-NaN wave height values
    hs_values = ds.hs.values
    assert not np.all(np.isnan(hs_values)), "All wave height values are NaN"

    ds.close()



"""Test lockup components."""

from copy import deepcopy

import pytest

# Import test utilities
from test_utils.logging import get_test_logger

# Initialize logger
logger = get_test_logger(__name__)

from rompy_swan.components.group import LOCKUP
from rompy_swan.components.lockup import (
    COMPUTE,
    COMPUTE_NONSTAT,
    COMPUTE_STAT,
    HOTFILE,
    STOP,
)
from rompy_swan.subcomponents.time import NONSTATIONARY, STATIONARY


@pytest.fixture(scope="module")
def times():
    yield NONSTATIONARY(
        tbeg="1990-01-01T00:00:00",
        tend="1990-02-01T00:00:00",
        delt="PT1H",
        tfmt=1,
        dfmt="hr",
    )


def test_compute():
    comp = COMPUTE()
    assert comp.render() == "COMPUTE"


def test_compute_stationary():
    comp = COMPUTE(times=dict(model_type="stationary", time="1990-01-01T00:00:00"))
    assert comp.render() == "COMPUTE STATIONARY time=19900101.000000"


def test_compute_nonstationary(times):
    comp = COMPUTE(times=times)
    assert comp.render() == (
        "COMPUTE NONSTATIONARY tbegc=19900101.000000 deltc=1.0 HR "
        "tendc=19900201.000000"
    )


def test_hotfile_default():
    hotfile = HOTFILE(fname="hotfile")
    assert hotfile.render() == "HOTFILE fname='hotfile'"


def test_hotfile_unformatted():
    hotfile = HOTFILE(fname="hotfile", format="unformatted")
    assert hotfile.render() == "HOTFILE fname='hotfile' UNFORMATTED"


def test_stop():
    stop = STOP()
    assert stop.render() == "STOP"


def test_compute_nonstationary_single_no_hotfiles(times):
    compute = COMPUTE_NONSTAT(times=times)
    assert compute.render() == f"COMPUTE {times.render()}"


def test_compute_nonstationary_single_with_hotfile(times):
    compute = COMPUTE_NONSTAT(times=times, hotfile={"fname": "hotfile"}, hottimes=[-1])
    cmds = compute.render().split("\n")
    assert len(cmds) == 2
    assert cmds[0].startswith("COMPUTE NONSTAT") and cmds[1].startswith("HOTFILE")


def test_compute_nonstationary_multiple_with_hotfiles(times):
    hottimes = ["1990-01-01T06:00:00", "1990-01-01T12:00:00", "1990-01-01T18:00:00"]
    compute = COMPUTE_NONSTAT(
        times=times,
        hotfile={"fname": "hotfile"},
        hottimes=hottimes,
    )
    cmds = compute.render().split("\n")
    assert len(cmds) == 7
    assert cmds[1] == "HOTFILE fname='hotfile_19900101T060000'"
    assert cmds[3] == "HOTFILE fname='hotfile_19900101T120000'"
    assert cmds[5] == "HOTFILE fname='hotfile_19900101T180000'"
    assert cmds[6].startswith("COMPUTE NONSTATIONARY tbegc=19900101.180000")


def test_compute_nonstationary_initstat(times):
    compute = COMPUTE_NONSTAT(times=times, initstat=True)
    cmds = compute.render().split("\n")
    assert len(cmds) == 2
    assert cmds[0].startswith("COMPUTE STATIONARY")
    assert cmds[1].startswith("COMPUTE NONSTATIONARY")


def test_compute_stationary_single_no_hotfile():
    compute = COMPUTE_STAT(times=STATIONARY(time="1990-01-01T00:00:00"))
    assert compute.render() == "COMPUTE STATIONARY time=19900101.000000"


def test_compute_stationary_single_with_hotfile(times):
    t = deepcopy(times)
    t.tend = t.tbeg
    compute = COMPUTE_STAT(times=t, hotfile=dict(fname="hotfile"), hottimes=[-1])
    cmds = compute.render().split("\n")
    assert len(cmds) == 2
    assert cmds[0] == "COMPUTE STATIONARY time=19900101.000000"
    assert cmds[1] == "HOTFILE fname='hotfile_19900101T000000'"


def test_compute_stationary_multiple_no_hotfiles(times):
    t = deepcopy(times)
    t.tend = times()[6]
    compute = COMPUTE_STAT(times=t)
    cmds = compute.render().split("\n")
    assert len(cmds) == 7
    for cmd in cmds:
        assert cmd.startswith("COMPUTE STATIONARY")


def test_compute_stationary_multiple_with_hotfiles(times):
    t = deepcopy(times)
    t.tend = times()[6]
    hottimes = ["1990-01-01T03:00:00", "1990-01-01T06:00:00"]
    compute = COMPUTE_STAT(times=t, hotfile={"fname": "hotfile"}, hottimes=hottimes)
    cmds = compute.render().split("\n")
    assert len(cmds) == 9
    for ind in [0, 1, 2, 3, 5, 6, 7]:
        assert cmds[ind].startswith("COMPUTE STATIONARY")
    for ind in [4, 8]:
        assert cmds[ind].startswith("HOTFILE")


def test_lockup_compute_stat(times):
    lockup = LOCKUP(
        compute=dict(
            model_type="stat",
            times=STATIONARY(time="1990-01-01T00:00:00"),
        ),
    )
    assert lockup.compute.model_type == "stat"
    assert "STOP" in lockup.render()


def test_lockup_compute_nonstat(times):
    lockup = LOCKUP(compute=dict(model_type="nonstat", times=times))
    assert lockup.compute.model_type == "nonstat"
    assert "STOP" in lockup.render()


def test_compute_stationary_no_times_passed():
    compute = COMPUTE_STAT()
    assert compute.render() == "COMPUTE STATIONARY time=19700101.000000"


def test_compute_nonstationary_no_times_passed():
    compute = COMPUTE_NONSTAT()
    assert compute.render() == f"COMPUTE {compute.times.render()}"



"""SWAN startup subcomponents."""

from typing import Literal

from pydantic import Field

from rompy_swan.subcomponents.base import BaseSubComponent


class CARTESIAN(BaseSubComponent):
    """Cartesian coordinates.

    .. code-block:: text

        CARTESIAN

    All locations and distances are in m. Coordinates are given with respect
    to x- and y-axes chosen by the user in the various commands.

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.components.startup import CARTESIAN
        coords = CARTESIAN()
        print(coords.render())

    """

    model_type: Literal["cartesian", "CARTESIAN"] = Field(
        default="cartesian", description="Model type discriminator"
    )


class SPHERICAL(BaseSubComponent):
    """Spherical coordinates.

    .. code-block:: text

        SPHERICAL [->CCM|QC]

    Notes
    -----

    projection options:

    * CCM: central conformal Mercator. The horizontal and vertical scales are
      uniform in terms of cm/degree over the area shown. In the centre of the scale
      is identical to that of the conventional Mercator projection (but only at
      that centre). The area in the projection centre is therefore exactly conformal.
    * QC: the projection method is quasi-cartesian, i.e. the horizontal and vertical
      scales are equal to one another in terms of cm/degree.

    All coordinates of locations and geographical grid sizes are given in degrees;`x`
    is longitude with `x = 0` being the Greenwich meridian and `x > 0` is East of this
    meridian; `y` is latitude with `y > 0` being the Northern hemisphere. Input and
    output grids have to be oriented with their x-axis to the East; mesh sizes are in
    degrees. All other distances are in meters.

    Note that spherical coordinates can also be used for relatively small areas, say 10
    or 20 km horizontal dimension. This may be useful if one obtains the boundary
    conditions by nesting in an oceanic model which is naturally formulated in
    spherical coordinates. Note that in case of spherical coordinates regular grids
    must always be oriented E-W, N-S, i.e. `alpc=0`, `alpinp=0`, `alpfr=0`
    (see commands CGRID, INPUT GRID and FRAME, respectively).

    Examples
    --------

    .. ipython:: python
        :okwarning:

        from rompy_swan.components.startup import SPHERICAL
        coords = SPHERICAL()
        print(coords.render())
        coords = SPHERICAL(projection="qc")
        print(coords.render())

    """

    model_type: Literal["spherical", "SPHERICAL"] = Field(
        default="spherical", description="Model type discriminator"
    )
    projection: Literal["ccm", "qc"] = Field(
        default="ccm",
        description=(
            "Defines the projection method in case of spherical coordinates, `ccm` "
            "Central Conformal Mercator, `qc` means Quasi-cartesian"
        ),
    )

    def cmd(self) -> str:
        """Render subcomponent cmd."""
        repr = super().cmd()
        if self.projection is not None:
            repr += f" {self.projection.upper()}"
        return repr



"""
SWAN Types

This module contains type definitions and enumerations used throughout the SWAN model
implementation, including grid types, boundary conditions, and physics options.
"""

from enum import Enum, IntEnum

from rompy.logging import get_logger

logger = get_logger(__name__)


class IDLA(IntEnum):
    """Order of values in the input files.

    Attributes
    ----------
    ONE: 1
        SWAN reads the map from left to right starting in the upper-left-hand corner of
        the map. A new line in the map should start on a new line in the file.
    TWO: 2
        As `1` but a new line in the map need not start on a new line in the file.
    THREE: 3
        SWAN reads the map from left to right starting in the lower-left-hand corner of
        the map. A new line in the map should start on a new line in the file.
    FOUR: 4
        As `3` but a new line in the map need not start on a new line in the file.
    FIVE: 5
        SWAN reads the map from top to bottom starting in the lower-left-hand corner of
        the map. A new column in the map should start on a new line in the file.
    SIX: 6
        As `5` but a new column in the map need not start on a new line in the file.

    Notes
    -----
    It is assumed that the x-axis of the grid is pointing to the right and the y-axis
    upwards.

    """

    ONE = 1
    TWO = 2
    THREE = 3
    FOUR = 4
    FIVE = 5
    SIX = 6


class GridOptions(str, Enum):
    """Valid options for the input grid type.

    Attributes
    ----------
    BOTTOM: "bottom"
        Bottom level grid.
    WLEVEL: "wlevel"
        Water level grid.
    CURRENT: "current"
        Current field grid.
    VX: "vx"
        Current field x-component grid.
    VY: "vy"
        Current field y-component grid.
    WIND: "wind"
        Wind velocity grid.
    WX: "wx"
        Wind velocity x-component grid.
    WY: "wy"
        Wind velocity y-component grid.
    FRICTION: "friction"
        Bottom friction grid.
    NPLANTS: "nplants"
        Horizontally varying vegetation density grid.
    TURBVISC: "turbvisc"
        Horizontally varying turbulent viscosity grid.
    MUDLAYER: "mudlayer"
        Horizontally varying mud layer thickness grid.
    AICE: "aice"
        Areal ice fraction grid, a number between 0 and 1.
    HICE: "hice"
        Ice thickness grid.
    HSS: "hss"
        Sea-swell significant wave height grid.
    TSS: "tss"
        Sea-swell mean wave period.

    """

    BOTTOM = "bottom"
    WLEVEL = "wlevel"
    CURRENT = "current"
    VX = "vx"
    VY = "vy"
    WIND = "wind"
    WX = "wx"
    WY = "wy"
    FRICTION = "friction"
    NPLANTS = "nplants"
    TURBVISC = "turbvisc"
    MUDLAYER = "mudlayer"
    AICE = "aice"
    HICE = "hice"
    HSS = "hss"
    TSS = "tss"


class BoundShapeOptions(str, Enum):
    """Valid options for the boundary shape type.

    Attributes
    ----------
    JONSWAP: "jonswap"
        JONSWAP spectrum.
    PM: "pm"
        Pierson-Moskowitz spectrum.
    GAUSS: "gauss"
        Gaussian spectrum.
    BIN: "bin"
        Energy at a single bin spectrum.
    TMA: "tma"
        TMA spectrum.

    """

    JONSWAP = "jonswap"
    PM = "pm"
    GAUSS = "gauss"
    BIN = "bin"
    TMA = "tma"


class SideOptions(str, Enum):
    """Valid options for the boundary shape type.

    Attributes
    ----------
    NORTH: "north"
        North side.
    NW: "nw"
        North-west side.
    WEST: "west"
        West side.
    SW: "sw"
        South-west side.
    SOUTH: "south"
        South side.
    SE: "se"
        South-east side.
    EAST: "east"
        East side.
    NE: "ne"
        North-east side.

    """

    NORTH = "north"
    NW = "nw"
    WEST = "west"
    SW = "sw"
    SOUTH = "south"
    SE = "se"
    EAST = "east"
    NE = "ne"


class PhysicsOff(str, Enum):
    """Physics commands to be switched off.

    Attributes
    ----------
    WINDGROWTH : str = "windgrowth"
        Switches off wind growth (in commands GEN1, GEN2, GEN3).
    QUADRUPL : str = "quadrupl"
        Switches off quadruplet wave interactions (in command GEN3).
    WCAPPING : str = "wcapping"
        Switches off whitecapping (in command GEN3).
    BREAKING : str = "breaking"
        Switches off wave breaking dissipation.
    REFRAC : str = "refrac"
        Switches off wave refraction (action transport in theta space).
    FSHIFT : str = "fshift"
        Switches off frequency shifting (action transport in sigma space).
    BNDCHK : str = "bndchk"
        Switches off the checking of the delta imposed and computed Hs at the boundary.

    """

    WINDGROWTH = "windgrowth"
    QUADRUPL = "quadrupl"
    WCAPPING = "wcapping"
    BREAKING = "breaking"
    REFRAC = "refrac"
    FSHIFT = "fshift"
    BNDCHK = "bndchk"


class BlockOptions(str, Enum):
    """Valid options for block output parameters.

    Attributes
    ----------
    HSIGN: "hsign"
        Significant wave height (in m).
    HSWELL: "hswell"
        Swell wave height (in m).
    DIR: "dir"
        Mean wave direction (in degrees).
    DPM: "dpm"
        Mean wave direction at the peak frequency (in degrees).
    PDIR: "pdir"
        Peak wave direction (in degrees).
    TDIR: "tdir"
        Direction of energy transport (in degrees).
    TM01: "tm01"
        Mean absolute wave period (in s).
    RTM01: "rtm01"
        Mean relative wave period (in s).
    RTP: "rtp"
        Peak period of the (relative frequency) variance density spectrum (in s).
    TPS: "tps"
        Smoothed peak period (in s).
    PER: "per"
        Mean absolute wave period (in s).
    RPER: "rper"
        Mean relative wave period (in s).
    TMM10: "tmm10"
        Mean absolute wave period (in s).
    RTMM10: "rtmm10"
        Mean relative wave period (in s).
    TM02: "tm02"
        Mean absolute zero-crossing period (in s).
    FSPR: "fspr"
        The normalised width of the frequency spectrum.
    DSPR: "dspr"
        Directional spreading of the waves (in degrees).
    QP: "qp"
        Peakedness of the wave spectrum (dimensionless).
    DEPTH: "depth"
        Water depth (in m).
    WATLEV: "watlev"
        Water level (in m).
    BOTLEV: "botlev"
        Bottom level (in m).
    VEL: "vel"
        Current velocity (vector; in m/s).
    FRCOEF: "frcoef"
        Friction coefficient (equal to `cfw` or `kn` in command `FRICTION`).
    WIND: "wind"
        Wind velocity (vector; in m/s).
    AICE: "aice"
        Ice concentration (as a fraction from 0 to 1).
    PROPAGAT: "propagat"
        Sum of PROPXY, PROPTHETA and PROPSIGMA (in W/m2 or m2/s).
    PROPXY: "propxy"
        Energy propagation in geographic space; sum of x- and y-direction
        (in W/m2 or m2/s).
    PROPTHETA: "proptheta"
        Energy propagation in theta space (in W/m2 or m2/s).
    PROPSIGMA: "propsigma"
        Energy propagation in sigma space (in W/m2 or m2/s).
    GENERAT: "generat"
        Total energy generation (in W/m2 or m2/s).
    GENWIND: "genwind"
        Energy generation due to wind (in W/m2 or m2/s).
    REDIST: "redist"
        Total energy redistribution (in W/m2 or m2/s).
    REDQUAD: "redquad"
        Energy redistribution due to quadruplets (in W/m2 or m2/s).
    REDTRIAD: "redtriad"
        Energy redistribution due to triads (in W/m2 or m2/s).
    DISSIP: "dissip"
        Total energy dissipation (in W/m2 or m2/s).
    DISBOT: "disbot"
        Energy dissipation due to bottom friction (in W/m2 or m2/s).
    DISSURF: "dissurf"
        Energy dissipation due to surf breaking (in W/m2 or m2/s).
    DISWCAP: "diswcap"
        Energy dissipation due to whitecapping (in W/m2 or m2/s).
    DISSWELL: "disswell"
        Energy dissipation due to swell dissipation (in W/m2 or m2/s).
    DISVEG: "disveg"
        Energy dissipation due to vegetation (in W/m2 or m2/s).
    DISMUD: "dismud"
        Energy dissipation due to mud (in W/m2 or m2/s).
    DISICE: "disice"
        Energy dissipation due to sea ice (in W/m2 or m2/s).
    RADSTR: "radstr"
        Energy transfer between waves and currents due to radiation stress
        (in W/m2 or m2/s).
    QB: "qb"
        Fraction of breaking waves due to depth-induced breaking.
    TRANSP: "transp"
        Transport of energy (vector; in W/m2 or m2/s).
    FORCE: "force"
        Wave-induced force per unit surface area (vector; in N/m2).
    UBOT: "ubot"
        The rms-value of the maxima of the orbital velocity near the bottom (in m/s).
    URMS: "urms"
        The rms-value of the orbital velocity near the bottom (in m/s).
    TMBOT: "tmbot"
        The bottom wave period (in s).
    WLENGTH: "wlength"
        Average wave length (in m).
    LWAVP: "lwavp"
        Peak wave length (in m).
    STEEPNESS: "steepness"
        Average wave steepness (dimensionless).
    BFI: "bfi"
        Benjamin-Feir index (dimensionless).
    NPLANTS: "nplants"
        Number of plants per square meter.
    DHSIGN: "dhsign"
        Difference in significant wave height from the last two iterations (in m).
    DRTM01: "drtm01"
        Difference in average wave period (RTM01) from the last two iterations (in s).
    LEAK: "leak"
        Numerical loss of energy equal to `cthetaE(omega,theta)` across boundaries.
    TIME: "time"
        Full date-time string as part of line used in TABLE only.
    TSEC: "tsec"
        Time in seconds with respect to a reference time (see command QUANTITY).
    XP: "xp"
        The x-coordinate in the problem coordinate system of the output location.
    YP: "yp"
        The y-coordinate in the problem coordinate system of the output location.
    DIST: "dist"
        If output has been requested along a curve then the distance along the curve
        can be obtained with the command TABLE. DIST is the distance along the curve
        measured from teh first point on the curve to the output location on the curve
        in meters (also in the case of spherical coordinates).
    SETUP: "setup"
        Set-up due to waves (in m).
    PTHSIGN: "pthsign"
        Watershed partitions of the significant wave height (in m).
    PTRTP: "ptrtp"
        Watershed partitions of the relative peak period (in s).
    PTWLEN: "ptwlen"
        Watershed partitions of the average wave length (in m).
    PTDIR: "ptdir"
        Watershed partitions of the peak wave direction (in degrees).
    PTDSPR: "ptdspr"
        Watershed partitions of the directional spreading (in degrees).
    PTWFRAC: "ptwfrac"
        Watershed partitions of the wind fraction (dimensionless).
    PTSTEEPNE: "ptsteepne"
        Watershed partition of the wave steepness (dimensionless).
    PARTITION: "partition"
        The raw spectral partition for wave system tracking post-processing.

    Note
    ----
    Energy given in W/m2 or m2/s depending on command SET.

    Note
    ----
    UBOT and URMS required command `FRICTION` to be used. If friction is ignored in the
    computation, then one should use the command `FRICTION` with the value of the
    friction set to zero (e.g., `FRICTION COLLINS 0`).

    """

    HSIGN = "hsign"
    HSWELL = "hswell"
    DIR = "dir"
    DPM = "dpm"
    PDIR = "pdir"
    TDIR = "tdir"
    TM01 = "tm01"
    RTM01 = "rtm01"
    RTP = "rtp"
    TPS = "tps"
    PER = "per"
    RPER = "rper"
    TMM10 = "tmm10"
    RTMM10 = "rtmm10"
    TM02 = "tm02"
    FSPR = "fspr"
    DSPR = "dspr"
    QP = "qp"
    DEPTH = "depth"
    WATLEV = "watlev"
    BOTLEV = "botlev"
    VEL = "vel"
    FRCOEF = "frcoef"
    WIND = "wind"
    AICE = "aice"
    PROPAGAT = "propagat"
    PROPXY = "propxy"
    PROPTHETA = "proptheta"
    PROPSIGMA = "propsigma"
    GENERAT = "generat"
    GENWIND = "genwind"
    REDIST = "redist"
    REDQUAD = "redquad"
    REDTRIAD = "redtriad"
    DISSIP = "dissip"
    DISBOT = "disbot"
    DISSURF = "dissurf"
    DISWCAP = "diswcap"
    DISSWELL = "disswell"
    DISVEG = "disveg"
    DISMUD = "dismud"
    DISICE = "disice"
    RADSTR = "radstr"
    QB = "qb"
    TRANSP = "transp"
    FORCE = "force"
    UBOT = "ubot"
    URMS = "urms"
    TMBOT = "tmbot"
    WLENGTH = "wlength"
    LWAVP = "lwavp"
    STEEPNESS = "steepness"
    BFI = "bfi"
    NPLANTS = "nplants"
    DHSIGN = "dhsign"
    DRTM01 = "drtm01"
    LEAK = "leak"
    TIME = "time"
    TSEC = "tsec"
    XP = "xp"
    YP = "yp"
    DIST = "dist"
    SETUP = "setup"
    PTHSIGN = "pthsign"
    PTRTP = "ptrtp"
    PTWLEN = "ptwlen"
    PTDIR = "ptdir"
    PTDSPR = "ptdspr"
    PTWFRAC = "ptwfrac"
    PTSTEEPNE = "ptsteepne"
    PARTITION = "partition"



__version__ = "0.6.0"
"""
SWAN Module for ROMPY

This module provides interfaces and utilities for working with the SWAN
(Simulating WAves Nearshore) model within the ROMPY framework.
"""

import os
from pathlib import Path

from rompy.logging import LoggingConfig, get_logger

logger = get_logger(__name__)

# Import SWAN components
from .boundary import Boundnest1
from .config import SwanConfig
from .data import SwanDataGrid
from .grid import SwanGrid

# Configure logging for the SWAN module
logging_config = LoggingConfig()
logging_config.configure_logging()

# Log module initialization
logger.debug("SWAN module initialized")



"""
SWAN Data Module

This module provides data handling functionality for the SWAN model within the ROMPY framework.
"""

import os
import time as time_module
from pathlib import Path
from typing import Optional

import numpy as np
import pandas as pd
import xarray as xr
from pydantic import Field, model_validator

from rompy.core.data import DataGrid
from rompy.core.time import TimeRange
from rompy.formatting import get_formatted_box, log_box
from rompy.logging import get_logger
from rompy_swan.grid import SwanGrid
from rompy_swan.types import GridOptions

logger = get_logger(__name__)

FILL_VALUE = -99.0


class SwanDataGrid(DataGrid):
    """This class is used to write SWAN data from a dataset."""

    z1: Optional[str] = Field(
        default=None,
        description=(
            "Name of the data variable in dataset representing either a scaler "
            "parameter or the u-componet of a vector field"
        ),
    )
    z2: Optional[str] = Field(
        default=None,
        description=(
            "Name of the data variable in dataset representing "
            "the v-componet of a vector field"
        ),
    )
    var: GridOptions = Field(description="SWAN input grid name")
    fac: float = Field(
        description=(
            "SWAN multiplies all values that are read from file by `fac`. For "
            "instance if the values are given in unit decimeter, one should make "
            "`fac=0.1` to obtain values in m. To change sign use a negative `fac`"
        ),
        default=1.0,
    )

    @model_validator(mode="after")
    def ensure_z1_in_data_vars(self) -> "SwanDataGrid":
        data_vars = self.variables
        for z in [self.z1, self.z2]:
            if z and z not in data_vars:
                logger.debug(f"Adding {z} to data_vars")
                data_vars.append(z)
        self.variables = data_vars
        return self

    def get(
        self,
        destdir: str | Path,
        grid: Optional[SwanGrid] = None,
        time: Optional[TimeRange] = None,
    ) -> Path:
        """Write the data source to a new location.

        Parameters
        ----------
        destdir : str | Path
            The destination directory to write the netcdf data to.
        grid: SwanGrid, optional
            The grid to filter the data to, only used if `self.filter_grid` is True.
        time: TimeRange, optional
            The times to filter the data to, only used if `self.filter_time` is True.

        Returns
        -------
        cmd: str
            The command line string with the INPGRID/READINP commands ready to be
            written to the SWAN input file.

        Note
        ----
        The data are assumed to not have been rotated. We cannot use the grid.rot attr
        as this is the rotation from the model grid object which is not necessarily the
        same as the rotation of the data.

        """
        if self.crop_data:
            if grid is not None:
                self._filter_grid(grid)
            if time is not None:
                self._filter_time(time)

        output_file = os.path.join(destdir, f"{self.var.value}.grd")

        # Create a formatted box for logging
        log_box(
            title=f"WRITING {self.var.value.upper()} GRID DATA",
            logger=logger,
            add_empty_line=False,
        )

        # Log output file and dataset information using bullet points
        items = [f"Output file: {output_file}"]

        # Add variable information if available
        if self.z1:
            shape_info = f"{self.ds[self.z1].shape}"
            items.append(f"Variable: {self.z1} with shape {shape_info}")
        if self.z2:
            shape_info = f"{self.ds[self.z2].shape}"
            items.append(f"Variable: {self.z2} with shape {shape_info}")

        # Add scaling factor
        items.append(f"Scaling factor: {self.fac}")

        # Log all items as a bulleted list
        logger.bullet_list(items, indent=2)

        start_time = time_module.time()
        if self.var.value == "bottom":
            inpgrid, readgrid = self.ds.swan.to_bottom_grid(
                output_file,
                fmt="%4.2f",
                x=self.coords.x,
                y=self.coords.y,
                z=self.z1,
                fac=self.fac,
                rot=0.0,
                vmin=float("-inf"),
            )
        else:
            inpgrid, readgrid = self.ds.swan.to_inpgrid(
                output_file=output_file,
                x=self.coords.x,
                y=self.coords.y,
                z1=self.z1,
                z2=self.z2,
                fac=self.fac,
                rot=0.0,
                var=self.var.name,
            )

        # Log completion and processing time
        elapsed_time = time_module.time() - start_time
        file_size = Path(output_file).stat().st_size / (1024 * 1024)  # Size in MB

        # Use the centralized functions from rompy package

        # Log completion information as a bulleted list
        logger.bullet_list(
            [
                f"Completed in {elapsed_time:.2f} seconds",
                f"File size: {file_size:.2f} MB",
            ],
            indent=2,
        )

        return f"{inpgrid}\n{readgrid}\n"

    def __str__(self):
        return f"SWANDataGrid {self.var.name}"

    def _format_value(self, obj):
        """Format SwanDataGrid values using the new formatting framework.

        This method provides special formatting for SwanDataGrid objects.

        Args:
            obj: The object to format

        Returns:
            A formatted string or None to use default formatting
        """
        # Only format SwanDataGrid objects
        if not isinstance(obj, SwanDataGrid):
            return None

        # Use the new formatting framework
        from rompy.formatting import format_value

        return format_value(obj)
        lines.append(f"  {bullet} Variable:   {obj.var.name}")

        # Add source information if available
        if hasattr(obj, "source") and obj.source:
            source_type = getattr(obj.source, "model_type", "unknown")
            lines.append(f"  {bullet} Source:     {source_type}")

            # Add dataset information if available
            if hasattr(obj.source, "dataset_id"):
                lines.append(f"  {bullet} Dataset ID: {obj.source.dataset_id}")

        # Add coordinate information if available
        if hasattr(obj, "coords") and obj.coords:
            coords = [f"{k}={v}" for k, v in obj.coords.items()]
            coords_str = ", ".join(coords)
            lines.append(f"  {bullet} Coordinates: {coords_str}")

        # Add scaling factor if available
        if hasattr(obj, "fac"):
            lines.append(f"  {bullet} Scale factor: {obj.fac}")

        # Add z variables information if available
        if hasattr(obj, "z1") and obj.z1:
            lines.append(f"  {bullet} Z1 variable: {obj.z1}")
        if hasattr(obj, "z2") and obj.z2:
            lines.append(f"  {bullet} Z2 variable: {obj.z2}")

        # Close with footer
        lines.append(footer)

        return "\n".join(lines)


def dset_to_swan(
    dset: xr.Dataset,
    output_file: str,
    variables: list,
    fmt: str = "%4.2f",
    fill_value: float = FILL_VALUE,
    time_dim="time",
):
    """Convert xarray Dataset into SWAN ASCII file.

    Parameters
    ----------
    dset: xr.Dataset
        Dataset to write in SWAN ASCII format.
    output_file: str
        Local file name for the ascii output file.
    variables: list
        Variables to write to ascii.
    fmt: str
        String float formatter.
    fill_value: float
        Fill value.
    time_dim: str
        Name of the time dimension if available in the dataset.

    """
    # Input checking
    for data_var in variables:
        if data_var not in dset.data_vars:
            raise ValueError(f"Variable {data_var} not in {dset}")
        if dset[data_var].ndim not in (2, 3):
            raise NotImplementedError(
                "Only 2D and 3D datasets are supported but "
                f"dset.{data_var} has {dset[data_var].ndim} dims"
            )

    # Ensure time is a dimension so iteration will work
    if time_dim not in dset.dims:
        dset = dset.expand_dims(time_dim, 0)

    # Write to ascii
    logger.debug(f"Writing SWAN ASCII file: {output_file}")

    # Use formatting utilities imported at the top of the file

    # Import formatting utilities at function level to avoid scoping issues
    # Create a formatted box for logging
    log_box(title="WRITING SWAN ASCII DATA", logger=logger)

    start_time = time_module.time()
    file_size = 0
    total_times = len(dset[time_dim])

    with open(output_file, "w") as stream:
        for i, t in enumerate(dset[time_dim]):
            time_str = pd.to_datetime(t.values)
            if (
                i % max(1, total_times // 10) == 0 or i == total_times - 1
            ):  # Log progress at 10% intervals
                logger.debug(
                    f"Writing progress: {i+1}/{total_times} times ({(i+1)/total_times*100:.1f}%) - Time: {time_str}"
                )
            else:
                logger.debug(f"Appending Time {time_str} to {output_file}")

            for data_var in variables:
                logger.debug(f"Appending Variable {data_var} to {output_file}")
                data = dset[data_var].sel(time=t).fillna(fill_value).values
                np.savetxt(fname=stream, X=data, fmt=fmt, delimiter="\t")

    elapsed_time = time_module.time() - start_time
    file_size = Path(output_file).stat().st_size / (1024 * 1024)  # Size in MB

    # Format the completion message
    elapsed_str = f"{elapsed_time:.2f}"
    size_str = f"{file_size:.2f}"
    # Get a formatted completion box
    completion_msg = f"COMPLETED: {elapsed_str} seconds, File size: {size_str} MB"
    completion_box = get_formatted_box(completion_msg)
    for line in completion_box.split("\n"):
        logger.debug(line)

    logger.debug(f"SWAN ASCII file written successfully to {output_file}")

    return output_file


@xr.register_dataset_accessor("swan")
class Swan_accessor(object):
    def __init__(self, xarray_obj):
        self._obj = xarray_obj

    def grid(
        self,
        x: str = "lon",
        y: str = "lat",
        rot: float = 0.0,
        exc: float = FILL_VALUE,
    ):
        """SWAN Grid object for this dataset.

        Parameters
        ----------
        x: str
            Name of the x coordinate variable.
        y: str
            Name of the y coordinate variable.
        rot: float
            Rotation angle in degrees, required if the grid has been previously rotated.
        exc: float
            Exception value for missing data.

        Returns
        -------
        grid: SwanGrid
            SwanGrid object representing this dataset.

        """
        return SwanGrid(
            grid_type="REG",
            x0=float(self._obj[x].min()),
            y0=float(self._obj[y].min()),
            dx=float(np.diff(self._obj[x]).mean()),
            dy=float(np.diff(self._obj[y]).mean()),
            nx=len(self._obj[x]),
            ny=len(self._obj[y]),
            rot=rot,
            exc=exc,
        )

    def to_bottom_grid(
        self,
        output_file,
        fmt="%4.2f",
        x="lon",
        y="lat",
        z="depth",
        fac=1.0,
        rot=0.0,
        vmin=float("-inf"),
        fill_value=FILL_VALUE,
    ):
        """Write SWAN inpgrid BOTTOM file.

        Parameters
        ----------
        output_file: str
            Local file name for the ascii output file.
        fmt: str
            String float formatter.
        x: str
            Name of the x-axis in the dataset.
        y: str
            Name of the y-axis in the dataset.
        z: str
            Name of the bottom variable in the dataset.
        rot: float
            Rotation angle, required if the grid has been previously rotated.
        vmin: float
            Minimum value below which depths are masked.
        fill_value: float
            Fill value.
        fac: float
            Multiplying factor in case data are not in m or should be reversed.

        Returns
        -------
        inpgrid: str
            SWAN INPgrid command instruction.
        readinp: str
            SWAN READinp command instruction.

        TODO: Merge this method with `to_inpgrid`.

        """
        dset_to_swan(
            dset=self._obj.where(self._obj > vmin).transpose(..., y, x),
            output_file=output_file,
            fmt=fmt,
            variables=[z],
            fill_value=fill_value,
        )
        grid = self.grid(x=x, y=y, rot=rot)
        inpgrid = f"INPGRID BOTTOM {grid.inpgrid}"
        readinp = f"READINP BOTTOM {fac} '{Path(output_file).name}' 3 FREE"
        return inpgrid, readinp

    def to_inpgrid(
        self,
        output_file: str,
        var: str = "WIND",
        fmt: str = "%.2f",
        x: str = "lon",
        y: str = "lat",
        z1: str = "u10",
        z2: str | None = None,
        fac: float = 1.0,
        rot: float = 0.0,
        time: str = "time",
    ):
        """This function writes to a SWAN inpgrid format file (i.e. WIND)

        Parameters
        ----------
        output_file: str
            Local file name for the ascii output file.
        var: str
            Type of swan input, used to define the INPGRID/READGRID strings.
        fmt: str
            String float formatter.
        x: str
            Name of the x-axis in the dataset.
        y: str
            Name of the y-axis in the dataset.
        z1: str
            Name of the first variable in the dataset.
        z2: str, optional
            Name of the second variable in the dataset.
        fac: float
            Multiplying factor in case data are not in m or should be reversed.
        rot: float
            Rotation angle, required if the grid has been previously rotated.
        time: str
            Name of the time variable in the dataset

        Returns
        -------
        inpgrid: str
            SWAN INPgrid command instruction.
        readinp: str
            SWAN READinp command instruction.

        """
        ds = self._obj

        # ds = ds.transpose((time,) + ds[x].dims)
        # Calculate time difference in hours
        time_diffs = np.diff(ds[time].values)
        dt = time_diffs.mean() / pd.to_timedelta(1, "h")
        dt_str = f"{dt:.2f}"  # Format as string to avoid formatting issues

        inptimes = []
        with open(output_file, "wt") as f:
            # iterate through time
            for ti, windtime in enumerate(ds[time].values):
                time_str = pd.to_datetime(windtime).strftime("%Y%m%d.%H%M%S")
                logger.debug(time_str)

                # write SWAN time header to file:
                f.write(f"{time_str}\n")

                # Write first component to file
                z1t = np.squeeze(ds[z1].isel(dict(time=ti)).values)
                np.savetxt(f, z1t, fmt=fmt)

                if z2 is not None:
                    z2t = np.squeeze(ds[z2].isel(dict(time=ti)).values)
                    np.savetxt(f, z2t, fmt=fmt)

                inptimes.append(time_str)

        if len(inptimes) < 1:
            os.remove(output_file)
            raise ValueError(
                f"***Error! No times written to {output_file}\n. Check the input data!"
            )

        # Create grid object from this dataset
        grid = self.grid(x=x, y=y, rot=rot)

        inpgrid = f"INPGRID {var} {grid.inpgrid} NONSTATION {inptimes[0]} {dt_str} HR"
        readinp = f"READINP {var} {fac} '{Path(output_file).name}' 3 0 1 0 FREE"

        # Log detailed information about the generated grid
        logger.debug(f"Created {var} grid with:")
        logger.debug(f"  → Grid size: {grid.nx}x{grid.ny} points")
        logger.debug(f"  → Resolution: dx={grid.dx}, dy={grid.dy}")
        logger.debug(f"  → Time points: {len(inptimes)}")
        logger.debug(f"  → Time interval: {dt_str} HR")

        return inpgrid, readinp

    def to_tpar_boundary(
        self,
        dest_path,
        boundary,
        interval,
        x_var="lon",
        y_var="lat",
        hs_var="sig_wav_ht",
        per_var="pk_wav_per",
        dir_var="pk_wav_dir",
        dir_spread=20.0,
    ):
        """This function writes parametric boundary forcing to a set of
        TPAR files at a given distance based on gridded wave output. It returns the string to be included in the Swan INPUT file.

        At present simple nearest neighbour point lookup is used.

        Args:
        TBD
        """
        from shapely.ops import substring

        bound_string = "BOUNDSPEC SEGM XY "
        point_string = "&\n {xp:0.8f} {yp:0.8f} "
        file_string = "&\n {len:0.8f} '{fname}' 1 "

        for xp, yp in boundary.exterior.coords:
            bound_string += point_string.format(xp=xp, yp=yp)

        bound_string += "&\n VAR FILE "

        n_pts = int((boundary.length) / interval)
        splits = np.linspace(0, 1.0, n_pts)
        j = 0
        for i in range(len(splits) - 1):
            segment = substring(
                boundary.exterior, splits[i], splits[i + 1], normalized=True
            )
            xp = segment.coords[1][0]
            yp = segment.coords[1][1]
            logger.debug(f"Extracting point: {xp},{yp}")
            ds_point = self._obj.sel(
                indexers={x_var: xp, y_var: yp}, method="nearest", tolerance=interval
            )
            if len(ds_point.time) == len(self._obj.time):
                if not np.any(np.isnan(ds_point[hs_var])):
                    output_tpar = f"{dest_path}/{j}.TPAR"
                    logger.debug(f"Writing boundary point {j} to {output_tpar}")
                    logger.debug(f"  → Location: ({xp:.5f}, {yp:.5f})")
                    logger.debug(f"  → Time points: {len(ds_point.time)}")

                    with open(output_tpar, "wt") as f:
                        f.write("TPAR\n")
                        for t in range(len(ds_point.time)):
                            ds_row = ds_point.isel(time=t)
                            lf = "{tt} {hs:0.2f} {per:0.2f} {dirn:0.1f} {spr:0.2f}\n"
                            f.write(
                                lf.format(
                                    tt=str(
                                        ds_row["time"]
                                        .dt.strftime("%Y%m%d.%H%M%S")
                                        .values
                                    ),
                                    hs=float(ds_row[hs_var]),
                                    per=float(ds_row[per_var]),
                                    dirn=float(ds_row[dir_var]),
                                    spr=dir_spread,
                                )
                            )
                    bound_string += file_string.format(
                        len=splits[i + 1] * boundary.length, fname=f"{j}.TPAR"
                    )
                    j += 1

        return bound_string



"""
SWAN Components Base Module

This module provides the base classes for SWAN components in the ROMPY framework.

How to subclass:
1. Define a new `model_type` Literal for the subclass
2. Overwrite the `cmd` method to return the SWAN input file string
"""

from abc import abstractmethod
from typing import Literal, Optional

from pydantic import ConfigDict, Field

from rompy.core.types import RompyBaseModel
from rompy.logging import get_logger

logger = get_logger(__name__)

# Constants
MAX_LENGTH = 180
SPACES = 4


def split_string(cmd: str, max_length: int = MAX_LENGTH, spaces: int = 4) -> list:
    """Split command cmd if longer than max_length.

    Parameters
    ----------
    cmd: str
        SWAN CMD string to split.
    max_length: int
        Maximum length of each split string.
    spaces: int
        Number of spaces in each identation before each new line.

    Returns
    -------
    split_cmd: list
        List of split CMD strings.

    """
    if len(cmd) <= max_length:
        return [cmd]

    split_index = cmd.rfind(" ", 0, max_length - 1 - spaces)

    if split_index == -1:
        split_index = max_length

    return [cmd[:split_index]] + split_string(cmd[split_index + 1 :])


class BaseComponent(RompyBaseModel):
    """Base class for SWAN components.

    This class is not intended to be used directly, but to be subclassed by other
    SWAN components to implement the following common behaviour:

    * Define a `render()` method to render a CMD string from the component
    * Split the rendered CMD if longer than 132 characters
    * Forbid extra arguments so only implemented fields must be specified

    """

    model_type: Literal["component"] = Field(description="Model type discriminator")
    model_config = ConfigDict(extra="forbid")

    def _render_split_cmd(self, cmd_line: str) -> str:
        """Split cmd_line if longer than MAX_LENGTH.

        Longer strings are recursively split by inserting a SWAN line continuation
        character `&` follwed by newline and identation until no line it soo long.

        Parameters
        ----------
        cmd_line: str
            Command line to split.

        Returns
        -------
        str
            Split command line.

        """
        # Split cmd at existing newlines
        cmd_lines = cmd_line.split("\n")
        # Split each line before max_length
        cmds = []
        for cmd in cmd_lines:
            cmds.extend(split_string(cmd, max_length=MAX_LENGTH, spaces=SPACES))
        # Joining lines
        return f" &\n{SPACES * ' '}".join(cmds)

    @abstractmethod
    def cmd(self) -> str | list:
        """Return the string or list of strings to render the component to the CMD."""
        pass

    def render(self, cmd: Optional[str | list] = None) -> str:
        """Render the component to a string.

        Parameters
        ----------
        cmd: Optional[str | list]
            Command string or list of command strings to render, by default self.cmd().

        Returns
        -------
        cdmstr: str
            The rendered command file component.

        """
        cmd_lines = cmd or self.cmd()
        if isinstance(cmd_lines, str):
            cmd_lines = [cmd_lines]
        cmd_lines = [self._render_split_cmd(cmd_line) for cmd_line in cmd_lines]
        return "\n".join(cmd_lines)


class MultiComponents(BaseComponent):
    """Mixin for components that can have multiple components."""

    components: list = Field(description="The components to render")

    def cmd(self) -> list[str]:
        repr = []
        for component in self.components:
            repr += [component.cmd()]
        return repr



"""Test startup components."""

# Import test utilities
from test_utils.logging import get_test_logger

# Initialize logger
logger = get_test_logger(__name__)

from rompy_swan.components.group import STARTUP
from rompy_swan.components.startup import COORDINATES, MODE, PROJECT, SET


def test_project():
    proj = PROJECT(
        name="Test project",
        nr="0001",
        title1="Title 1",
        title2="Title 2",
        title3="Title 3",
    )
    assert proj.render() == (
        "PROJECT name='Test project' nr='0001' "
        "title1='Title 1' title2='Title 2' title3='Title 3'"
    )


def test_project_no_titles():
    proj = PROJECT(
        name="Test project",
        nr="0001",
    )
    assert proj.render() == "PROJECT name='Test project' nr='0001'"


def test_set():
    s = SET(direction_convention="nautical")
    assert s.render() == "SET NAUTICAL"


def test_mode_default():
    mode = MODE()
    assert mode.render() == "MODE STATIONARY TWODIMENSIONAL"


def test_mode_non_default():
    mode = MODE(kind="stationary", dim="onedimensional")
    assert mode.render() == "MODE STATIONARY ONEDIMENSIONAL"


def test_coord_default():
    coord = COORDINATES()
    assert coord.render() == "COORDINATES CARTESIAN"


def test_startup():
    proj = PROJECT(nr="01")
    set = SET(level=0.5, direction_convention="nautical")
    mode = MODE(kind="nonstationary", dim="twodimensional")
    coords = COORDINATES(kind=dict(model_type="spherical", projection="ccm"))
    startup = STARTUP(
        project=proj,
        set=set,
        mode=mode,
        coordinates=coords,
    )
    assert "PROJECT nr='01'" in startup.render()
    assert "SET level=0.5 NAUTICAL" in startup.render()
    assert "MODE NONSTATIONARY TWODIMENSIONAL" in startup.render()
    assert "COORDINATES SPHERICAL CCM" in startup.render()



# NOTE: This file previously used PySchism modules which have been removed.
# Consider updating to use PyLibs equivalents or other alternatives.
# PySchism imports have been commented out - functionality may be limited.
import logging
from datetime import datetime
from pathlib import Path
from time import time

import xarray as xr
from matplotlib.transforms import Bbox
# from pyschism.forcing.hycom.hycom2schism import DownloadHycom  # TODO: Replace with PyLibs equivalent
# from pyschism.mesh.hgrid import Hgrid  # TODO: Replace with PyLibs equivalent

"""
Download hycom data for Fortran scripts.
Default is to download data for generating initial condition (use hgrid 
    as parameter to cover the whole domain).
Optionally, download data for generating th.nc (bnd=True) and nu.nc (nudge=True) 
    (use bbox as parameter to cut small domain).
"""
logging.basicConfig(
    format="[%(asctime)s] %(name)s %(levelname)s: %(message)s",
    force=True,
)
logger = logging.getLogger("pyschism")
logger.setLevel(logging.INFO)


hgrid = Hgrid.open("./hgrid.gr3", crs="epsg:4326")
date = datetime(2023, 1, 2)
hycom = DownloadHycom(hgrid=hgrid)
t0 = time()
hycom.fetch_data(date, rnday=2, bnd=False, nudge=False, sub_sample=5, outdir="./")
print(f"It took {(time()-t0)/60} mins to download")

files = Path("./").glob("hycom_*.nc")
xr.open_mfdataset(files, concat_dim="time", combine="nested").to_netcdf("hycom.nc")
print(
    "run ncrcat  hycom_*.nc hycom.nc to combine daily output files into compatible input"
)



"""Test swan_types."""

import pytest

# Import test utilities
from test_utils.logging import get_test_logger

# Initialize logger
logger = get_test_logger(__name__)

from rompy_swan.types import IDLA, BoundShapeOptions, GridOptions, SideOptions


def test_idla():
    assert IDLA(1) == IDLA.ONE == 1
    assert IDLA(2) == IDLA.TWO == 2
    assert IDLA(3) == IDLA.THREE == 3
    assert IDLA(4) == IDLA.FOUR == 4
    assert IDLA(5) == IDLA.FIVE == 5
    assert IDLA(6) == IDLA.SIX == 6
    with pytest.raises(ValueError):
        IDLA(7)


def test_grid_options():
    assert GridOptions.BOTTOM.name == "BOTTOM"
    assert GridOptions.BOTTOM.value == "bottom"
    assert GridOptions.WLEVEL.name == "WLEVEL"
    assert GridOptions.WLEVEL.value == "wlevel"
    assert GridOptions.CURRENT.name == "CURRENT"
    assert GridOptions.CURRENT.value == "current"
    assert GridOptions.VX.name == "VX"
    assert GridOptions.VX.value == "vx"
    assert GridOptions.VY.name == "VY"
    assert GridOptions.VY.value == "vy"
    assert GridOptions.WIND.name == "WIND"
    assert GridOptions.WIND.value == "wind"
    assert GridOptions.WX.name == "WX"
    assert GridOptions.WX.value == "wx"
    assert GridOptions.WY.name == "WY"
    assert GridOptions.WY.value == "wy"
    assert GridOptions.FRICTION.name == "FRICTION"
    assert GridOptions.FRICTION.value == "friction"
    assert GridOptions.NPLANTS.name == "NPLANTS"
    assert GridOptions.NPLANTS.value == "nplants"
    assert GridOptions.TURBVISC.name == "TURBVISC"
    assert GridOptions.TURBVISC.value == "turbvisc"
    assert GridOptions.MUDLAYER.name == "MUDLAYER"
    assert GridOptions.MUDLAYER.value == "mudlayer"
    assert GridOptions.AICE.name == "AICE"
    assert GridOptions.AICE.value == "aice"
    assert GridOptions.HICE.name == "HICE"
    assert GridOptions.HICE.value == "hice"
    assert GridOptions.HSS.name == "HSS"
    assert GridOptions.HSS.value == "hss"
    assert GridOptions.TSS.name == "TSS"
    assert GridOptions.TSS.value == "tss"
    for option in GridOptions:
        assert option.name == option.value.upper()
    with pytest.raises(ValueError):
        GridOptions("invalid")


def test_bound_shape_options():
    assert BoundShapeOptions.JONSWAP.name == "JONSWAP"
    assert BoundShapeOptions.JONSWAP.value == "jonswap"
    assert BoundShapeOptions.PM.name == "PM"
    assert BoundShapeOptions.PM.value == "pm"
    assert BoundShapeOptions.GAUSS.name == "GAUSS"
    assert BoundShapeOptions.GAUSS.value == "gauss"
    assert BoundShapeOptions.BIN.name == "BIN"
    assert BoundShapeOptions.BIN.value == "bin"
    assert BoundShapeOptions.TMA.name == "TMA"
    assert BoundShapeOptions.TMA.value == "tma"
    for option in BoundShapeOptions:
        assert option.name == option.value.upper()
    with pytest.raises(ValueError):
        BoundShapeOptions("square")


def test_side_options():
    assert SideOptions.NORTH.name == "NORTH"
    assert SideOptions.NORTH.value == "north"
    assert SideOptions.NW.name == "NW"
    assert SideOptions.NW.value == "nw"
    assert SideOptions.WEST.name == "WEST"
    assert SideOptions.WEST.value == "west"
    assert SideOptions.SW.name == "SW"
    assert SideOptions.SW.value == "sw"
    assert SideOptions.SOUTH.name == "SOUTH"
    assert SideOptions.SOUTH.value == "south"
    assert SideOptions.SE.name == "SE"
    assert SideOptions.SE.value == "se"
    assert SideOptions.EAST.name == "EAST"
    assert SideOptions.EAST.value == "east"
    assert SideOptions.NE.name == "NE"
    assert SideOptions.NE.value == "ne"
    for option in SideOptions:
        assert option.name == option.value.upper()
    with pytest.raises(ValueError):
        SideOptions("center")



"""Test time sub-component."""

import pytest

# Import test utilities
from test_utils.logging import get_test_logger

# Initialize logger
logger = get_test_logger(__name__)

from datetime import datetime, timedelta

from rompy_swan.subcomponents.time import (
    NONSTATIONARY,
    STATIONARY,
    Delt,
    Time,
    TimeRangeClosed,
    TimeRangeOpen,
)


@pytest.mark.parametrize(
    "tfmt, tmpl",
    [
        (1, "%Y%m%d.%H%M%S"),
        (2, "'%d-%b-%y %H:%M:%S'"),
        (3, "%m/%d/%y.%H:%M:%S"),
        (4, "%H:%M:%S"),
        (5, "%y/%m/%d %H:%M:%S'"),
        (6, "%y%m%d%H%M"),
    ],
)
def test_time(tfmt, tmpl):
    t = datetime(1990, 1, 1, 0, 0, 0)
    time = Time(time=t, tfmt=tfmt)
    assert time.render() == t.strftime(tmpl)


@pytest.mark.parametrize(
    "delt, dfmt, rendered",
    [
        (1800, "sec", "1800.0 SEC"),
        (timedelta(hours=2), "min", "120.0 MIN"),
        ("PT60M", "hr", "1.0 HR"),
    ],
)
def test_delt(delt, dfmt, rendered):
    delta = Delt(delt=delt, dfmt=dfmt)
    assert delta.render() == rendered


def test_timerange_open():
    tr = TimeRangeOpen(tbeg="2023-01-01T00:00:00", delt="PT30M", dfmt="min")
    assert tr.render() == "tbeg=20230101.000000 delt=30.0 MIN"


def test_timerange_closed():
    tr = TimeRangeClosed(
        tbeg="2023-01-01T00:00:00",
        tend="2023-02-01T00:00:00",
        delt="PT30M",
        dfmt="min",
    )
    assert tr.render() == "tbeg=20230101.000000 delt=30.0 MIN tend=20230201.000000"


def test_nonstationary():
    tr = NONSTATIONARY(
        tbeg="2023-01-01T00:00:00",
        tend="2023-02-01T00:00:00",
        delt="PT30M",
        tfmt=6,
        dfmt="sec",
        suffix="tbl",
    )
    assert tr.render() == (
        "NONSTATIONARY tbegtbl=2301010000 delttbl=1800.0 SEC tendtbl=2302010000"
    )


def test_stationary():
    stat = STATIONARY(time="2023-01-01T00:00:00", tfmt=1)
    assert stat.render() == "STATIONARY time=20230101.000000"
